import functools
import logging
from typing import Optional

from cryptography import x509, exceptions
from cryptography.hazmat.primitives.asymmetric import (
    padding, rsa, dsa, ec, ed25519, ed448
)
from pyasn1.codec.der.encoder import encode
from pyasn1_alt_modules import rfc5280, rfc3739

from pkilint import validation, pkix, document
from pkilint.document import Document, ValueDecoder
from pkilint.itu.bitstring import NamedBitStringMinimalEncodingValidator
from pkilint.itu.string import PrintableStringConstraintValidator
from pkilint.pkix import (extension, time, name,
                          create_name_validator_container, general_name, algorithm
                          )
from pkilint.pkix.certificate import (
    certificate_validity, certificate_extension, certificate_validator,
    certificate_key, certificate_name, certificate_transparency,
)

logger = logging.getLogger(__name__)


class RFC5280Certificate(Document):
    def __init__(self, substrate_source, substrate,
                 name=None, parent=None
                 ):
        super().__init__(
            rfc5280.Certificate(), substrate_source, substrate, name, parent
        )

    @property
    def not_before(self):
        return time.parse_time_node(
            self.root.navigate('tbsCertificate.validity.notBefore')
        )

    @property
    def not_after(self):
        return time.parse_time_node(
            self.root.navigate('tbsCertificate.validity.notAfter')
        )

    @property
    def is_ca(self) -> Optional[bool]:
        ext_and_idx = self.get_extension_by_oid(rfc5280.id_ce_basicConstraints)

        if ext_and_idx is None:
            return False

        ext, _ = ext_and_idx

        try:
            return bool(ext.navigate('extnValue.basicConstraints.cA').pdu)
        except document.PDUNavigationFailedError:
            return None

    @functools.cached_property
    def cryptography_object(self):
        return x509.load_der_x509_certificate(self.substrate)

    @functools.cached_property
    def is_self_issued(self):
        issuer_node = self.root.navigate('tbsCertificate.issuer')
        subject_node = self.root.navigate('tbsCertificate.subject')

        return encode(issuer_node.pdu) == encode(subject_node.pdu)

    @functools.cached_property
    def is_self_signed(self):
        if not self.is_self_issued:
            return False

        public_key = self.cryptography_object.public_key()
        tbs_octets = self.cryptography_object.tbs_certificate_bytes
        hash_alg = self.cryptography_object.signature_hash_algorithm
        signature_octets = self.cryptography_object.signature

        try:
            if isinstance(public_key, rsa.RSAPublicKey):
                public_key.verify(
                    signature_octets,
                    tbs_octets,
                    padding.PKCS1v15(),
                    hash_alg
                )
            elif isinstance(public_key, dsa.DSAPublicKey):
                public_key.verify(
                    signature_octets,
                    tbs_octets,
                    hash_alg
                )
            elif isinstance(public_key, ec.EllipticCurvePublicKey):
                public_key.verify(
                    signature_octets,
                    tbs_octets,
                    ec.ECDSA(hash_alg)
                )
            elif isinstance(public_key, ed25519.Ed25519PublicKey):
                public_key.verify(
                    signature_octets,
                    tbs_octets
                )
            elif isinstance(public_key, ed448.Ed448PublicKey):
                public_key.verify(
                    signature_octets,
                    tbs_octets
                )
        except exceptions.InvalidSignature:
            logger.exception('Signature validation failed')

            return False

        return True

    def get_extension_by_oid(self, oid):
        tbs_cert = self.root.children['tbsCertificate']

        # ensure there's extensions
        if 'extensions' not in tbs_cert.children:
            return None

        extensions_node = tbs_cert.children['extensions']

        for ext_idx, extension_node in extensions_node.children.items():
            ext_oid = extension_node.children['extnID'].pdu

            if ext_oid == oid:
                return extension_node, int(ext_idx)

        return None

    def get_name_attributes_by_type(self, oid, name_path):
        name_node = self.root.navigate(name_path)

        return name.get_name_attributes_by_type(name_node, oid)

    def get_issuer_attributes_by_type(self, oid):
        return self.get_name_attributes_by_type(oid, 'tbsCertificate.issuer')

    def get_subject_attributes_by_type(self, oid):
        return self.get_name_attributes_by_type(oid, 'tbsCertificate.subject')

    def has_policy_oid(self, oid, inhibit_anypolicy=False):
        ext = self.get_extension_by_oid(rfc5280.id_ce_certificatePolicies)

        if ext is None:
            return False

        ext_value, _ = ext

        oids = [
            pi.children['policyIdentifier'].pdu
            for pi in (
                ext_value.navigate(
                    'extnValue.certificatePolicies'
                ).children.values()
            )
        ]

        if not inhibit_anypolicy and rfc5280.anyPolicy in oids:
            return True

        return oid in oids


def create_spki_decoder(subject_public_key_type_mappings, subject_public_key_parameters_type_mappings):
    subject_public_key_decoder = certificate_key.SubjectPublicKeyDecoder(
        type_mappings=subject_public_key_type_mappings
    )
    subject_public_key_parameters_decoder = certificate_key.SubjectPublicKeyParametersDecoder(
        type_mappings=subject_public_key_parameters_type_mappings
    )

    return validation.ValidatorContainer(
        validators=[
            certificate_key.SubjectPublicKeyDecodingValidator(
                decode_func=subject_public_key_decoder
            ),
            certificate_key.SubjectPublicKeyParametersDecodingValidator(
                decode_func=subject_public_key_parameters_decoder
            )
        ],
        pdu_class=rfc5280.SubjectPublicKeyInfo
    )


def create_policy_qualifier_decoder(type_mappings):
    decoder = ValueDecoder(
        type_path='policyQualifierId',
        value_path='qualifier',
        type_mappings=type_mappings
    )

    return validation.DecodingValidator(decode_func=decoder,
                                        pdu_class=rfc5280.PolicyQualifierInfo
                                        )


def create_other_name_decoder(type_mappings):
    decoder = ValueDecoder(
        type_path='type-id',
        value_path='value',
        type_mappings=type_mappings
    )

    return validation.DecodingValidator(decode_func=decoder,
                                        pdu_class=rfc5280.AnotherName
                                        )


def create_qc_statements_decoder(type_mappings):
    decoder = ValueDecoder(
        type_path='statementId',
        value_path='statementInfo',
        type_mappings=type_mappings
    )

    return validation.DecodingValidator(decode_func=decoder, pdu_class=rfc3739.QCStatement)


def create_issuer_validator_container(additional_validators=None, **kwargs):
    if additional_validators is None:
        additional_validators = []

    if len(kwargs) == 0:
        kwargs['path'] = 'certificate.tbsCertificate.issuer'

    return create_name_validator_container([
                                               name.EmptyNameValidator(),
                                               general_name.MailboxAddressSyntaxValidator(
                                                   pdu_class=rfc5280.EmailAddress
                                               )
                                           ] + additional_validators,
                                           **kwargs
                                           )


def create_subject_validator_container(additional_validators=None, **kwargs):
    if additional_validators is None:
        additional_validators = []

    if len(kwargs) == 0:
        kwargs['path'] = 'certificate.tbsCertificate.subject'

    return create_name_validator_container([
                                               certificate_name.SubjectEmailAddressInSanValidator(),
                                               general_name.MailboxAddressSyntaxValidator(
                                                   pdu_class=rfc5280.EmailAddress
                                               ),
                                               name.DomainComponentValidDomainNameValidator(
                                                   pdu_class=rfc5280.Name
                                               ),
                                           ] + additional_validators,
                                           **kwargs
                                           )


def create_validity_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
                       certificate_validity.CertificateSaneValidityPeriodValidator(),
                       time.TimeCorrectEncodingValidator(),
                   ] + additional_validators,
        path='certificate.tbsCertificate.validity'
    )


def create_extensions_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
                       extension.UniqueExtensionValidator(),
                       certificate_extension.BasicConstraintsValidator(),
                       certificate_extension.CertificatePolicyQualifierValidator(),
                       certificate_extension.CertificatePoliciesUserNoticeValidator(),
                       certificate_key.SubjectKeyIdentifierValidator(),
                       certificate_extension.SubjectKeyIdentifierCriticalityValidator(),
                       certificate_extension.KeyUsageCriticalityValidator(),
                       certificate_extension.KeyUsageValidator(),
                       general_name.UriSyntaxValidator(pdu_class=rfc5280.CPSuri),
                       general_name.GeneralNameUriSyntaxValidator(),
                       general_name.GeneralNameDnsNameSyntaxValidator(),
                       general_name.GeneralNameIpAddressSyntaxValidator(),
                       general_name.GeneralNameMailboxAddressSyntaxValidator(),
                       general_name.SmtpUTF8MailboxValidator(),
                       certificate_extension.DuplicatePolicyValidator(),
                       certificate_extension.CrlDpCriticalityValidator(),
                       certificate_extension.NameConstraintsCriticalityValidator(),
                       extension.AuthorityKeyIdentifierValidator(),
                       extension.AuthorityKeyIdentifierCriticalityValidator(),
                       certificate_extension.NameConstraintsValidator(),
                       certificate_extension.NameConstraintsGeneralSubtreeValidator(),
                       certificate_extension.AuthorityInformationAccessCriticalityValidator(),
                       certificate_extension.SubjectInformationAccessCriticalityValidator(),
                       certificate_extension.SubjectAlternativeNameCriticalityValidator(),
                       certificate_extension.SubjectDirectoryAttributesCriticalityValidator(),
                       certificate_extension.SmimeCapabilitiesCriticalityValidator(),
                       extension.DistributionPointValidator(),
                       certificate_extension.CtPrecertPoisonCriticalityValidator(),
                       certificate_extension.CtPrecertPoisonSctListMutuallyExclusiveExtensionsValidator(),
                       certificate_transparency.SctListElementCountValidator(),
                   ] + additional_validators,
        path='certificate.tbsCertificate.extensions'
    )


def create_pkix_certificate_validator_container(
        decoding_validators, validators):
    decoding_validator_container = [
        validation.ValidatorContainer(
            validators=decoding_validators, path='certificate'
        )
    ]

    validators += [
        PrintableStringConstraintValidator(),
        certificate_validator.CorrectVersionValidator(),
        pkix.CertificateSerialNumberValidator(),
        certificate_validator.SignatureAlgorithmMatchValidator(),
        certificate_extension.SubjectKeyIdentifierPresenceValidator(),
        certificate_extension.AuthorityKeyIdentifierPresenceValidator(),
        certificate_extension.KeyUsagePresenceValidator(),
        time.UtcTimeCorrectSyntaxValidator(),
        time.GeneralizedTimeCorrectSyntaxValidator(),
        NamedBitStringMinimalEncodingValidator(),
        certificate_validator.IssuerUniqueIdAbsenceValidator(),
        certificate_validator.SubjectUniqueIdAbsenceValidator(),
    ]

    return validation.ValidatorContainer(
        validators=decoding_validator_container + validators
    )


def create_decoding_validators(name_mappings, extension_mappings, additional_decoding_validators=None):
    if additional_decoding_validators is None:
        additional_decoding_validators = []
    return [
        pkix.create_attribute_decoder(
            name_mappings
        ),
        pkix.create_extension_decoder(
            extension_mappings
        ),
        pkix.create_signature_algorithm_identifier_decoder(
            algorithm.SIGNATURE_ALGORITHM_IDENTIFIER_MAPPINGS,
            path='certificate.tbsCertificate.signature'
        ),
        create_spki_decoder(
            certificate_key.SUBJECT_PUBLIC_KEY_ALGORITHM_IDENTIFIER_MAPPINGS,
            certificate_key.SUBJECT_KEY_PARAMETER_ALGORITHM_IDENTIFIER_MAPPINGS
        ),
        create_policy_qualifier_decoder(
            certificate_extension.CERTIFICATE_POLICY_QUALIFIER_MAPPINGS
        ),
        create_other_name_decoder(
            general_name.OTHER_NAME_MAPPINGS
        ),
        certificate_transparency.SctListExtensionDecodingValidator(),
    ] + additional_decoding_validators



import operator

from dateutil.relativedelta import relativedelta
from pyasn1_alt_modules import rfc5280, rfc6962

import pkilint.common
from pkilint import validation, document
from pkilint.pkix import time, Rfc2119Word


_CODE_CLASSIFIER = 'cabf.serverauth.root'


class RootExtensionAllowanceValidator(pkilint.common.ExtensionIdentifierAllowanceValidator):
    """Validates that the included extensions conform to BR 7.1.2.1.2."""

    _EXTENSION_ALLOWANCES = {
        rfc5280.id_ce_authorityKeyIdentifier: Rfc2119Word.SHOULD,
        rfc5280.id_ce_basicConstraints: Rfc2119Word.MUST,
        rfc5280.id_ce_keyUsage: Rfc2119Word.MUST,
        rfc5280.id_ce_subjectKeyIdentifier: Rfc2119Word.MUST,
        rfc5280.id_ce_extKeyUsage: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_certificatePolicies: Rfc2119Word.SHOULD_NOT,
        rfc6962.id_ce_embeddedSCT: Rfc2119Word.MAY
    }

    def __init__(self):
        super().__init__(self._EXTENSION_ALLOWANCES, _CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


class RootSubjectIssuerIdenticalEncodingValidator(validation.DEREqualityValidator):
    """Validates that the encoding of the subject and issuer DN are identical, as per BR 7.1.2.1."""

    # BR 7.1.2.1
    VALIDATION_ROOT_SUBJECT_ISSUER_DN_ENCODING_NOT_IDENTICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.root_subject_issuer_name_encoding_not_equal'
    )

    def __init__(self):
        super().__init__(other_node_retriever=lambda n: n.navigate('^.issuer'),
                         validation=self.VALIDATION_ROOT_SUBJECT_ISSUER_DN_ENCODING_NOT_IDENTICAL,
                         path='certificate.tbsCertificate.subject')


class RootAkiSkiEqualityValidator(validation.Validator):
    """Validates that the key identifier as encoded in the subject key identifier and authority key identifier
    extensions is identical, as per BR 7.1.2.1.3."""

    VALIDATION_ROOT_SKI_AKI_NOT_EQUAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.root_aki_ski_not_equal'
    )

    @staticmethod
    def _get_aki_value_node(node):
        aki_and_idx = node.document.get_extension_by_oid(rfc5280.id_ce_authorityKeyIdentifier)

        if aki_and_idx is None:
            return None

        aki, _ = aki_and_idx

        try:
            return aki.navigate('extnValue.authorityKeyIdentifier.keyIdentifier')
        except document.PDUNavigationFailedError:
            return None

    def __init__(self):
        super().__init__(validations=self.VALIDATION_ROOT_SKI_AKI_NOT_EQUAL,
                         pdu_class=rfc5280.SubjectKeyIdentifier)

    def validate(self, node):
        aki_value_node = RootAkiSkiEqualityValidator._get_aki_value_node(node)

        if aki_value_node is None:
            return

        ski_octets = node.pdu.asOctets().hex()
        aki_octets = aki_value_node.pdu.asOctets().hex()

        if ski_octets != aki_octets:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ROOT_SKI_AKI_NOT_EQUAL,
                f'SKI octets: "{ski_octets}", AKI octets: "{aki_octets}"')


class RootBasicConstraintsValidator(validation.Validator):
    """Validates that the content of the basic constraints extension conforms to BR 7.1.2.1.4."""

    VALIDATION_BC_CA_NOT_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.root_basic_constraints_ca_not_present'
    )

    VALIDATION_BC_PATHLEN_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.root_basic_constraints_pathlenconstraint_present'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_BC_CA_NOT_PRESENT, self.VALIDATION_BC_PATHLEN_PRESENT],
                         pdu_class=rfc5280.BasicConstraints)

    def validate(self, node):
        if not bool(node.children['cA'].pdu):
            raise validation.ValidationFindingEncountered(self.VALIDATION_BC_CA_NOT_PRESENT)

        if 'pathLenConstraint' in node.children:
            raise validation.ValidationFindingEncountered(self.VALIDATION_BC_PATHLEN_PRESENT)


class RootValidityPeriodValidator(time.ValidityPeriodThresholdsValidator):
    """Validates that the validity period conforms with BR 7.1.2.1.1."""

    VALIDATION_ROOT_VALIDITY_PERIOD_TOO_SHORT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.root_validity_period_too_short'
    )

    VALIDATION_ROOT_VALIDITY_PERIOD_TOO_LONG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.root_validity_period_too_long'
    )

    def __init__(self):
        minimum_threshold = (operator.ge, relativedelta(days=2922), self.VALIDATION_ROOT_VALIDITY_PERIOD_TOO_SHORT)
        maximum_threshold = (operator.le, relativedelta(days=9132), self.VALIDATION_ROOT_VALIDITY_PERIOD_TOO_LONG)

        super().__init__(end_validity_node_retriever=lambda n: n.navigate('^.notAfter'),
                         inclusive_second=True,
                         validity_period_thresholds=[minimum_threshold, maximum_threshold],
                         path='certificate.tbsCertificate.validity.notBefore'
                         )



import enum

from pyasn1.type.univ import ObjectIdentifier

BR_VERSION = '2.0.0'


ID_POLICY_EV = ObjectIdentifier('2.23.140.1.1')
ID_POLICY_DV = ObjectIdentifier('2.23.140.1.2.1')
ID_POLICY_OV = ObjectIdentifier('2.23.140.1.2.2')
ID_POLICY_IV = ObjectIdentifier('2.23.140.1.2.3')

SERVERAUTH_RESERVED_POLICY_OIDS = {ID_POLICY_EV, ID_POLICY_DV, ID_POLICY_OV, ID_POLICY_IV}


@enum.unique
class CertificateType(enum.IntEnum):
    ROOT_CA = 1,
    INTERNAL_CROSS_CA = 2,
    EXTERNAL_CROSS_CA = 3,
    NON_TLS_CA = 4,
    PRECERT_SIGNING_CA = 5,
    INTERNAL_UNCONSTRAINED_TLS_CA = 6,
    INTERNAL_CONSTRAINED_TLS_CA = 7,
    EXTERNAL_UNCONSTRAINED_TLS_CA = 8,
    EXTERNAL_UNCONSTRAINED_EV_TLS_CA = 9,
    EXTERNAL_CONSTRAINED_TLS_CA = 10,
    EXTERNAL_CONSTRAINED_EV_TLS_CA = 11,
    DV_FINAL_CERTIFICATE = 12,
    IV_FINAL_CERTIFICATE = 13,
    OV_FINAL_CERTIFICATE = 14,
    EV_FINAL_CERTIFICATE = 15,
    OCSP_RESPONDER = 16,
    DV_PRE_CERTIFICATE = 17,
    IV_PRE_CERTIFICATE = 18,
    OV_PRE_CERTIFICATE = 19,
    EV_PRE_CERTIFICATE = 20,

    def __str__(self):
        return self.name

    @property
    def to_option_str(self):
        return self.name.replace('_', '-')

    @staticmethod
    def from_option_str(value):
        value = value.replace('-', '_').upper()

        return CertificateType[value]


INTERMEDIATE_CERTIFICATE_TYPES = {
    CertificateType.INTERNAL_CROSS_CA,
    CertificateType.EXTERNAL_CROSS_CA,
    CertificateType.NON_TLS_CA,
    CertificateType.PRECERT_SIGNING_CA,
    CertificateType.INTERNAL_UNCONSTRAINED_TLS_CA,
    CertificateType.INTERNAL_CONSTRAINED_TLS_CA,
    CertificateType.EXTERNAL_UNCONSTRAINED_TLS_CA,
    CertificateType.EXTERNAL_UNCONSTRAINED_EV_TLS_CA,
    CertificateType.EXTERNAL_CONSTRAINED_TLS_CA,
    CertificateType.EXTERNAL_CONSTRAINED_EV_TLS_CA,
}

CROSS_CA_TYPES = {
    CertificateType.INTERNAL_CROSS_CA,
    CertificateType.EXTERNAL_CROSS_CA
}

INTERNAL_CA_TYPES = {
    CertificateType.INTERNAL_CROSS_CA,
    CertificateType.INTERNAL_UNCONSTRAINED_TLS_CA,
    CertificateType.INTERNAL_CONSTRAINED_TLS_CA,
    CertificateType.NON_TLS_CA,
}

EXTERNAL_CA_TYPES = {
    CertificateType.EXTERNAL_UNCONSTRAINED_TLS_CA,
    CertificateType.EXTERNAL_CONSTRAINED_EV_TLS_CA,
    CertificateType.EXTERNAL_UNCONSTRAINED_EV_TLS_CA,
    CertificateType.EXTERNAL_CONSTRAINED_TLS_CA,
    CertificateType.EXTERNAL_CROSS_CA,
}

CONSTRAINED_TLS_CA_TYPES = {
    CertificateType.EXTERNAL_CONSTRAINED_EV_TLS_CA,
    CertificateType.EXTERNAL_CONSTRAINED_TLS_CA,
    CertificateType.INTERNAL_CONSTRAINED_TLS_CA,
}

TLS_CA_TYPES = {
                   CertificateType.INTERNAL_CROSS_CA,
                   CertificateType.INTERNAL_UNCONSTRAINED_TLS_CA,
                   CertificateType.INTERNAL_CONSTRAINED_TLS_CA,
               } | EXTERNAL_CA_TYPES

SUBSCRIBER_FINAL_CERTIFICATE_TYPES = {
    CertificateType.DV_FINAL_CERTIFICATE,
    CertificateType.IV_FINAL_CERTIFICATE,
    CertificateType.OV_FINAL_CERTIFICATE,
    CertificateType.EV_FINAL_CERTIFICATE,
}

SUBSCRIBER_PRECERT_TYPES = {
    CertificateType.DV_PRE_CERTIFICATE,
    CertificateType.IV_PRE_CERTIFICATE,
    CertificateType.OV_PRE_CERTIFICATE,
    CertificateType.EV_PRE_CERTIFICATE,
}

SUBSCRIBER_CERTIFICATE_TYPES = SUBSCRIBER_FINAL_CERTIFICATE_TYPES | SUBSCRIBER_PRECERT_TYPES

DV_CERTIFICATE_TYPES = {CertificateType.DV_PRE_CERTIFICATE, CertificateType.DV_FINAL_CERTIFICATE}
IV_CERTIFICATE_TYPES = {CertificateType.IV_PRE_CERTIFICATE, CertificateType.IV_FINAL_CERTIFICATE}
OV_CERTIFICATE_TYPES = {CertificateType.OV_PRE_CERTIFICATE, CertificateType.OV_FINAL_CERTIFICATE}
EV_CERTIFICATE_TYPES = {CertificateType.EV_PRE_CERTIFICATE, CertificateType.EV_FINAL_CERTIFICATE}

IDENTITY_CERTIFICATE_TYPES = IV_CERTIFICATE_TYPES | OV_CERTIFICATE_TYPES | EV_CERTIFICATE_TYPES



from pyasn1_alt_modules import rfc3279, rfc5480, rfc5280

from pkilint import validation
from pkilint.pkix.certificate import certificate_key


class RsaKeyValidator(validation.Validator):
    VALIDATION_RSA_MODULUS_INVALID_LENGTH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.rsa_modulus_invalid_length'
    )

    VALIDATION_RSA_MODULUS_HAS_SMALL_PRIME_FACTOR = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.rsa_modulus_has_small_prime_factor'
    )

    VALIDATION_RSA_EXPONENT_PROHIBITED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.rsa_exponent_prohibited_value'
    )

    VALIDATION_RSA_EXPONENT_NOT_IN_RECOMMENDED_RANGE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.rsa_exponent_not_in_recommended_range'
    )

    _PRIMES_UNDER_752 = {2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97,
                         101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193,
                         197, 199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263, 269, 271, 277, 281, 283, 293, 307,
                         311, 313, 317, 331, 337, 347, 349, 353, 359, 367, 373, 379, 383, 389, 397, 401, 409, 419, 421,
                         431, 433, 439, 443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 503, 509, 521, 523, 541, 547,
                         557, 563, 569, 571, 577, 587, 593, 599, 601, 607, 613, 617, 619, 631, 641, 643, 647, 653, 659,
                         661, 673, 677, 683, 691, 701, 709, 719, 727, 733, 739, 743, 751}

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_RSA_MODULUS_INVALID_LENGTH,
                self.VALIDATION_RSA_MODULUS_HAS_SMALL_PRIME_FACTOR,
                self.VALIDATION_RSA_EXPONENT_PROHIBITED,
                self.VALIDATION_RSA_EXPONENT_NOT_IN_RECOMMENDED_RANGE,
            ],
            pdu_class=rfc3279.RSAPublicKey
        )

    def validate(self, node):
        modulus = int(node.children['modulus'].pdu)
        modulus_len = modulus.bit_length()
        exponent = int(node.children['publicExponent'].pdu)

        results = []

        if modulus_len < 2048 or modulus_len % 8 != 0:
            results.append(validation.ValidationFindingDescription(
                self.VALIDATION_RSA_MODULUS_INVALID_LENGTH,
                f'Invalid modulus length: {modulus_len}'
            ))

        small_prime = next(
            filter(
                lambda p: modulus % p == 0,
                self._PRIMES_UNDER_752
            ),
            None
        )

        if small_prime is not None:
            results.append(validation.ValidationFindingDescription(
                self.VALIDATION_RSA_MODULUS_HAS_SMALL_PRIME_FACTOR,
                f'Modulus has prime factor less than 752: {small_prime}'
            ))

        if exponent % 2 == 0:
            results.append(validation.ValidationFindingDescription(
                self.VALIDATION_RSA_EXPONENT_PROHIBITED,
                f'Exponent is not odd: {exponent}'
            ))
        elif exponent < 3:
            results.append(validation.ValidationFindingDescription(
                self.VALIDATION_RSA_EXPONENT_PROHIBITED,
                f'Exponent is less than 3: {exponent}'
            ))

        if exponent < ((1 << 16) + 1) or exponent > ((1 << 256) - 1):
            results.append(validation.ValidationFindingDescription(
                self.VALIDATION_RSA_EXPONENT_NOT_IN_RECOMMENDED_RANGE,
                f'Exponent is out of recommended range: {exponent}'
            ))

        return validation.ValidationResult(self, node, results)


class EcdsaKeyValidator(validation.Validator):
    VALIDATION_KEY_VALIDATION_FAILED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.ecdsa_key_validation_failed'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_KEY_VALIDATION_FAILED],
            pdu_class=rfc5280.SubjectPublicKeyInfo,
            predicate=lambda n: n.navigate('algorithm.algorithm').pdu == rfc5480.id_ecPublicKey
        )

    def validate(self, node):
        try:
            key_obj = certificate_key.convert_spki_to_object(node)

            if key_obj is None:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_KEY_VALIDATION_FAILED,
                    'Unsupported key type'
                )
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_KEY_VALIDATION_FAILED,
                str(e)
            )



from typing import NamedTuple, Set

import unicodedata
from pyasn1.type.univ import ObjectIdentifier
from pyasn1_alt_modules import rfc5280, rfc4262, rfc6962

from pkilint import validation, oid
from pkilint.itu.bitstring import has_named_bit
from pkilint.pkix import extension
from pkilint.pkix.extension import (get_criticality_from_decoded_node,
                                    ExtensionCriticalityValidator
                                    )

CERTIFICATE_POLICY_QUALIFIER_MAPPINGS = {
    rfc5280.id_qt_cps: rfc5280.CPSuri(),
    rfc5280.id_qt_unotice: rfc5280.UserNotice(),
}


class BasicConstraintsValidator(validation.Validator):
    VALIDATION_ILLEGAL_PATHLEN_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.basic_constraints_has_pathlen_for_non_ca'
    )
    VALIDATION_NOT_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.basic_constraints_extension_not_critical'
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.BasicConstraints,
            validations=[
                self.VALIDATION_ILLEGAL_PATHLEN_SET,
                self.VALIDATION_NOT_CRITICAL,
            ]
        )

    def validate(self, node):
        is_ca = bool(node.children['cA'].pdu)

        if not is_ca and 'pathLenConstraint' in node.children:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ILLEGAL_PATHLEN_SET,
                f'Certificate is end-entity but has pathLenConstraint of '
                f'{int(node.children["pathLenConstraint"].pdu)} set'
            )

        is_critical = get_criticality_from_decoded_node(node)

        if not is_critical and is_ca:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NOT_CRITICAL,
                'CA certificate does not have critical basicConstraints'
            )


class CertificatePolicySet(NamedTuple):
    required: bool
    policies: Set[ObjectIdentifier]


class CertificatePolicyOIDValidator(validation.Validator):
    VALIDATION_REQUIRED_POLICY_OID_NOT_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.required_certificate_policy_oid_missing'
    )

    VALIDATION_CONFLICTING_POLICY_OIDS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.conflicting_certificate_policy_oids'
    )

    VALIDATION_UNKNOWN_POLICY_OID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.unknown_certificate_policy_oid'
    )

    VALIDATION_ANYPOLICY_IN_END_ENTITY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.anypolicy_certificate_policy_in_end_entity_certificate'
    )

    def __init__(self, *, policy_sets, allow_other_policies=True):
        self._policy_sets = policy_sets
        self._allow_other_policies = allow_other_policies

        validations = [
            self.VALIDATION_REQUIRED_POLICY_OID_NOT_PRESENT,
            self.VALIDATION_CONFLICTING_POLICY_OIDS,
            self.VALIDATION_ANYPOLICY_IN_END_ENTITY,
        ]

        if not allow_other_policies:
            validations.append(self.VALIDATION_UNKNOWN_POLICY_OID)

        super().__init__(
            pdu_class=rfc5280.CertificatePolicies,
            validations=validations
        )

    def validate(self, node):
        policy_oids = {
            policy.children['policyIdentifier'].pdu
            for policy in node.children.values()
        }

        findings = []

        if not self._allow_other_policies:
            all_specified_policies = set()
            for policy_set in self._policy_sets:
                all_specified_policies += policy_set.policies

            other_policies = policy_oids - all_specified_policies
            if len(other_policies) > 0:
                other_policies_str = oid.format_oids(other_policies)
                findings.append(validation.ValidationFindingDescription(
                    self.VALIDATION_UNKNOWN_POLICY_OID,
                    f'Unknown certificate policies: {other_policies_str}'
                ))

        is_ca = node.document.is_ca

        for policy_set in self._policy_sets:
            policy_intersection = policy_set.policies & policy_oids

            if policy_set.required and len(policy_intersection) == 0:
                policies_str = oid.format_oids(policy_set.policies)

                findings.append(validation.ValidationFindingDescription(
                    self.VALIDATION_REQUIRED_POLICY_OID_NOT_PRESENT,
                    f'None of the following certificate policies are present: '
                    f'{policies_str}'
                ))

            if not is_ca and len(policy_intersection) > 1:
                policies_str = oid.format_oids(policy_intersection)

                findings.append(validation.ValidationFindingDescription(
                    self.VALIDATION_CONFLICTING_POLICY_OIDS,
                    f'Conflicting certificate policies present: {policies_str}'
                ))

        if not is_ca and rfc5280.anyPolicy in policy_oids:
            findings.append(validation.ValidationFindingDescription(
                self.VALIDATION_ANYPOLICY_IN_END_ENTITY,
                None
            ))

        return validation.ValidationResult(self, node, findings)


class SubjectKeyIdentifierPresenceValidator(validation.Validator):
    VALIDATION_EE_SKID_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.certificate_skid_end_entity_missing'
    )

    VALIDATION_CA_SKID_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.certificate_skid_ca_missing'
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.Certificate,
            validations=[
                self.VALIDATION_EE_SKID_MISSING,
                self.VALIDATION_CA_SKID_MISSING,
            ])

    def validate(self, node):
        is_ca = node.document.is_ca

        skid_ext = node.document.get_extension_by_oid(
            rfc5280.id_ce_subjectKeyIdentifier
        )

        if skid_ext is None:
            if is_ca:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_CA_SKID_MISSING
                )
            else:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_EE_SKID_MISSING
                )


class SubjectKeyIdentifierCriticalityValidator(ExtensionCriticalityValidator):
    VALIDATION_SKID_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.certificate_skid_extension_critical'
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc5280.id_ce_subjectKeyIdentifier,
            is_critical=False,
            validation=self.VALIDATION_SKID_CRITICAL
        )


class CrlDpCriticalityValidator(ExtensionCriticalityValidator):
    VALIDATION_CRLDP_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.certificate_crldp_extension_critical'
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc5280.id_ce_cRLDistributionPoints,
            is_critical=False,
            validation=self.VALIDATION_CRLDP_CRITICAL
        )


class NameConstraintsCriticalityValidator(ExtensionCriticalityValidator):
    VALIDATION_NC_NOT_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.certificate_name_constraints_extension_not_critical'
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc5280.id_ce_nameConstraints,
            is_critical=True,
            validation=self.VALIDATION_NC_NOT_CRITICAL
        )


class NameConstraintsValidator(validation.Validator):
    VALIDATION_NO_SUBTREES = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.name_constraints_no_subtrees'
    )

    VALIDATION_NC_IN_EE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.name_constraints_in_ee_certificate'
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_NO_SUBTREES,
                self.VALIDATION_NC_IN_EE
            ],
            pdu_class=rfc5280.NameConstraints
        )

    def validate(self, node):
        if not node.document.is_ca:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NC_IN_EE
            )

        if len(node.children) == 0:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NO_SUBTREES
            )


class NameConstraintsGeneralSubtreeValidator(validation.Validator):
    VALIDATION_NON_DEFAULT_MINIMUM = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.name_constraints_non_default_minimum'
    )

    VALIDATION_MAXIMUM_SPECIFIED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.name_constraints_maximum_specified'
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_NON_DEFAULT_MINIMUM,
                self.VALIDATION_MAXIMUM_SPECIFIED,
            ],
            pdu_class=rfc5280.GeneralSubtree
        )

    def validate(self, node):
        results = []

        minimum_dist = int(node.children['minimum'].pdu)
        if minimum_dist != 0:
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_NON_DEFAULT_MINIMUM,
                    f'Non-default minimum distance: {minimum_dist}'
                )
            )

        if 'maximum' in node.children:
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_MAXIMUM_SPECIFIED,
                    None
                )
            )

        return validation.ValidationResult(self, node, results)


def _get_policy_oids(ext):
    return {
        pi.children['policyIdentifier'].pdu
        for pi in ext.children.values()
    }


class IssuerSubjectPolicyChainValidator(validation.Validator):
    VALIDATION_SUBJECT_CERTIFICATE_POLICY_MISMATCH = (
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.NOTICE,
            'pkix.certificate_subject_has_policy_not_in_issuer'
        )
    )

    VALIDATION_SUBJECT_NO_CP_EXTENSION = validation.ValidationFinding(
        validation.ValidationFindingSeverity.INFO,
        'pkix.certificate_has_no_certificate_policies_extension'
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.CertificatePolicies,
            validations=[
                self.VALIDATION_SUBJECT_CERTIFICATE_POLICY_MISMATCH,
                self.VALIDATION_SUBJECT_NO_CP_EXTENSION,
            ]
        )

    def validate(self, node):
        subject_cert = node.document.parent['subject']

        cp_ext_and_idx = subject_cert.get_extension_by_oid(
            rfc5280.id_ce_certificatePolicies
        )

        if cp_ext_and_idx is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SUBJECT_NO_CP_EXTENSION
            )

        subject_cert_cp_ext_decoded = cp_ext_and_idx[0].navigate(
            'extnValue.certificatePolicies'
        )

        issuer_cert_policy_oids = _get_policy_oids(node)
        subject_cert_policy_oids = _get_policy_oids(
            subject_cert_cp_ext_decoded
        )

        if rfc5280.anyPolicy not in issuer_cert_policy_oids:
            difference = subject_cert_policy_oids - issuer_cert_policy_oids

            if len(difference) > 0:
                difference_str = ', '.join(map(str, difference))

                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_SUBJECT_CERTIFICATE_POLICY_MISMATCH,
                    f'Policy OIDs found in subject certificate but not in '
                    f'issuer certificate: {difference_str}'
                )


class DuplicatePolicyValidator(validation.Validator):
    VALIDATION_DUPLICATE_POLICIES = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.duplicate_certificate_policy_oids'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_DUPLICATE_POLICIES],
                         pdu_class=rfc5280.CertificatePolicies
                         )

    def validate(self, node):
        policy_oids = [
            pi.children['policyIdentifier'].pdu
            for pi in node.children.values()
        ]

        duplicates = [
            o
            for o in policy_oids
            if policy_oids.count(o) > 1
        ]

        if len(duplicates) > 0:
            dups_str = ', '.join(map(str, duplicates))

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_DUPLICATE_POLICIES,
                f'Duplicate policy identifiers: {dups_str}'
            )


class CertificatePolicyQualifierValidator(validation.Validator):
    VALIDATION_POLICY_HAS_QUALIFIER = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'pkix.certificate_policies_policy_has_qualifier'
    )

    VALIDATION_ANYPOLICY_DISALLOWED_QUALIFIER = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.certificate_policies_anypolicy_has_prohibited_qualifier'
    )

    def __init__(self):
        super().__init__(pdu_class=rfc5280.PolicyInformation,
                         validations=[
                             self.VALIDATION_POLICY_HAS_QUALIFIER,
                             self.VALIDATION_ANYPOLICY_DISALLOWED_QUALIFIER,
                         ]
                         )

    def validate(self, node):
        if 'policyQualifiers' in node.children:
            if node.children['policyIdentifier'].pdu == rfc5280.anyPolicy:
                disallowed_qualifiers = [
                    q.children['policyQualifierId'].pdu
                    for q in node.children['policyQualifiers'].children.values()
                    if q.children['policyQualifierId'].pdu not in [
                        rfc5280.id_qt_cps,
                        rfc5280.id_qt_unotice,
                    ]
                ]

                if len(disallowed_qualifiers) > 0:
                    message = ', '.join(map(str, disallowed_qualifiers))

                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_ANYPOLICY_DISALLOWED_QUALIFIER,
                        f'anyPolicy has disallowed qualifiers: {message}'
                    )

            if len(node.children['policyQualifiers'].children) > 0:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_POLICY_HAS_QUALIFIER
                )


class CertificatePoliciesUserNoticeValidator(validation.Validator):
    VALIDATION_NOTICEREF_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.certificate_policies_usernotice_has_noticeRef'
    )

    VALIDATION_EXPLICITTEXT_INVALID_ENCODING_5280 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.rfc5280_certificate_policies_invalid_explicit_text_encoding'
    )

    VALIDATION_EXPLICITTEXT_INVALID_ENCODING_6818 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.rfc6818_certificate_policies_invalid_explicit_text_encoding'
    )

    VALIDATION_EXPLICITTEXT_HAS_CONTROL_CHARACTER = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.certificate_policies_explicit_text_has_control_character'
    )

    VALIDATION_EXPLICITTEXT_NOT_NFC = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.certificate_policies_explicit_text_not_nfc_normalized'
    )

    def __init__(self):
        super().__init__(pdu_class=rfc5280.UserNotice, validations=[
            self.VALIDATION_NOTICEREF_PRESENT,
            self.VALIDATION_EXPLICITTEXT_INVALID_ENCODING_5280,
            self.VALIDATION_EXPLICITTEXT_INVALID_ENCODING_6818,
            self.VALIDATION_EXPLICITTEXT_HAS_CONTROL_CHARACTER,
            self.VALIDATION_EXPLICITTEXT_NOT_NFC,
        ])

    def validate(self, node):
        results = []

        if 'noticeRef' in node.children:
            results.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_NOTICEREF_PRESENT,
                    None
                )
            )

        explicit_text = node.children.get('explicitText')

        if explicit_text is not None:
            encoding, value = explicit_text.child

            if encoding not in ['ia5String', 'utf8String']:
                results.append(
                    validation.ValidationFindingDescription(
                        self.VALIDATION_EXPLICITTEXT_INVALID_ENCODING_5280,
                        f'Invalid encoding: {encoding}'
                    )
                )

            if encoding not in ['bmpString', 'utf8String', 'visibleString']:
                results.append(
                    validation.ValidationFindingDescription(
                        self.VALIDATION_EXPLICITTEXT_INVALID_ENCODING_6818,
                        f'Invalid encoding: {encoding}'
                    )
                )

            if str(value.pdu) != unicodedata.normalize('NFC', str(value.pdu)):
                results.append(
                    validation.ValidationFindingDescription(
                        self.VALIDATION_EXPLICITTEXT_NOT_NFC,
                        None
                    )
                )

            if any((
                    ord(c) <= 0x1f or (0x7f <= ord(c) <= 0x9f)
                    for c in str(value.pdu)
            )):
                results.append(
                    validation.ValidationFindingDescription(
                        self.VALIDATION_EXPLICITTEXT_HAS_CONTROL_CHARACTER,
                        None
                    )
                )

        return validation.ValidationResult(self, node, results)


class KeyUsagePresenceValidator(extension.ExtensionPresenceValidator):
    VALIDATION_CA_NO_KU_EXTENSION = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ca_certificate_no_ku_extension'
    )

    def __init__(self):
        super().__init__(
            extension_oid=rfc5280.id_ce_keyUsage,
            validation=self.VALIDATION_CA_NO_KU_EXTENSION,
            pdu_class=rfc5280.Extensions,
            predicate=lambda n: n.document.is_ca
        )


class KeyUsageCriticalityValidator(validation.Validator):
    VALIDATION_KU_NOT_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.key_usage_extension_not_critical'
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.KeyUsage,
            validations=[self.VALIDATION_KU_NOT_CRITICAL]
        )

    def validate(self, node):
        if not get_criticality_from_decoded_node(node):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_KU_NOT_CRITICAL
            )


class KeyUsageValidator(validation.Validator):
    VALIDATION_NO_BITS_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.no_ku_bits_set'
    )

    VALIDATION_CA_KEYCERTSIGN_NOT_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ca_certificate_keycertsign_keyusage_not_set'
    )

    VALIDATION_EE_KEYCERTSIGN_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ee_certificate_keycertsign_keyusage_set'
    )

    VALIDATION_EO_AND_DO_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.both_encipheronly_and_decipheronly_ku_set'
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.KeyUsage,
            validations=[
                self.VALIDATION_NO_BITS_SET,
                self.VALIDATION_CA_KEYCERTSIGN_NOT_SET,
                self.VALIDATION_EE_KEYCERTSIGN_SET,
                self.VALIDATION_EO_AND_DO_SET,
            ]
        )

    def validate(self, node):
        if not any(
                map(
                    lambda kn: has_named_bit(node, kn),
                    rfc5280.KeyUsage.namedValues.keys()
                )
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NO_BITS_SET
            )

        is_ca = node.document.is_ca

        has_keycertsign = has_named_bit(node, 'keyCertSign')

        if is_ca and not has_keycertsign:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CA_KEYCERTSIGN_NOT_SET
            )
        elif not is_ca and has_keycertsign:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_EE_KEYCERTSIGN_SET
            )

        has_eo = has_named_bit(node, 'encipherOnly')
        has_do = has_named_bit(node, 'decipherOnly')

        if has_eo and has_do:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_EO_AND_DO_SET
            )


class AuthorityKeyIdentifierPresenceValidator(validation.Validator):
    VALIDATION_AKI_EXTENSION_NOT_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.authority_key_identifier_extension_absent'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_AKI_EXTENSION_NOT_PRESENT],
            pdu_class=rfc5280.Certificate
        )

    def validate(self, node):
        ext = node.document.get_extension_by_oid(
            rfc5280.id_ce_authorityKeyIdentifier
        )

        if ext is None and not node.document.is_self_signed:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AKI_EXTENSION_NOT_PRESENT
            )


class AuthorityInformationAccessCriticalityValidator(
    ExtensionCriticalityValidator
):
    VALIDATION_AIA_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.authority_information_access_extension_critical'
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc5280.id_pe_authorityInfoAccess,
            is_critical=False,
            validation=self.VALIDATION_AIA_EXTENSION_CRITICAL
        )


class AuthorityKeyIdentifierCriticalityValidator(ExtensionCriticalityValidator):
    VALIDATION_AKID_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.certificate_akid_extension_critical'
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc5280.id_ce_authorityKeyIdentifier,
            is_critical=False,
            validation=self.VALIDATION_AKID_CRITICAL
        )


class SubjectInformationAccessCriticalityValidator(
    ExtensionCriticalityValidator
):
    VALIDATION_SIA_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.subject_information_access_extension_critical'
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc5280.id_pe_subjectInfoAccess,
            is_critical=False,
            validation=self.VALIDATION_SIA_EXTENSION_CRITICAL
        )


class SubjectAlternativeNameCriticalityValidator(validation.TypeMatchingValidator):
    VALIDATION_SAN_NOT_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.san_extension_not_critical_empty_subject'
    )

    VALIDATION_SAN_IS_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.san_extension_is_critical_non_empty_subject'
    )

    def __init__(self):
        super().__init__(type_path='extnID', type_oid=rfc5280.id_ce_subjectAltName, value_path='critical',
                         pdu_class=rfc5280.Extension, validations=[
                            self.VALIDATION_SAN_NOT_CRITICAL, self.VALIDATION_SAN_IS_CRITICAL
                         ])

    def validate_with_value(self, node, value_node):
        is_critical = bool(value_node.pdu)

        subject_node = node.navigate(':certificate.tbsCertificate.subject.rdnSequence')

        if not any(subject_node.children) and not is_critical:
            raise validation.ValidationFindingEncountered(self.VALIDATION_SAN_NOT_CRITICAL)

        if any(subject_node.children) and is_critical:
            raise validation.ValidationFindingEncountered(self.VALIDATION_SAN_IS_CRITICAL)


class SubjectDirectoryAttributesCriticalityValidator(ExtensionCriticalityValidator):
    VALIDATION_SDA_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.subject_directory_attributes_extension_critical'
    )

    def __init__(self):
        super().__init__(type_oid=rfc5280.id_ce_subjectDirectoryAttributes, is_critical=False,
                         validation=self.VALIDATION_SDA_EXTENSION_CRITICAL)


class SmimeCapabilitiesCriticalityValidator(ExtensionCriticalityValidator):
    VALIDATION_SMIME_CAPABILITIES_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.smime_capabilities_extension_critical'
    )

    def __init__(self):
        super().__init__(type_oid=rfc4262.smimeCapabilities, is_critical=False,
                         validation=self.VALIDATION_SMIME_CAPABILITIES_EXTENSION_CRITICAL)


class CtPrecertPoisonCriticalityValidator(ExtensionCriticalityValidator):
    VALIDATION_CT_PRECERT_POISON_NOT_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ct_precert_poison_extension_not_critical'
    )

    def __init__(self):
        super().__init__(type_oid=rfc6962.id_ce_criticalPoison, is_critical=True,
                         validation=self.VALIDATION_CT_PRECERT_POISON_NOT_CRITICAL)


class CtPrecertPoisonSctListMutuallyExclusiveExtensionsValidator(validation.Validator):
    VALIDATION_SIMULTANEOUS_PRECERT_POISON_SCTLIST_EXTENSIONS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ct_precert_poison_and_sctlist_extensions_present'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_SIMULTANEOUS_PRECERT_POISON_SCTLIST_EXTENSIONS,
                         pdu_class=rfc6962.SignedCertificateTimestampList)

    def validate(self, node):
        poison_ext = node.document.get_extension_by_oid(rfc6962.id_ce_criticalPoison)

        if poison_ext is not None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SIMULTANEOUS_PRECERT_POISON_SCTLIST_EXTENSIONS)



#!/usr/bin/env python

import argparse
import sys

from pkilint import pkix
from pkilint import util, loader, report
from pkilint.pkix import extension, name, ocsp


def main(cli_args=None) -> int:
    parser = argparse.ArgumentParser(description='RFC 6960 OCSP Response Linter')

    subparsers = parser.add_subparsers(dest='command', required=True)
    subparsers.add_parser('validations', help='Output the set of validations which this linter performs')

    lint_parser = subparsers.add_parser('lint', help='Lint the specified OCSP response')
    util.add_standard_args(lint_parser)

    lint_parser.add_argument('file', type=argparse.FileType('rb'),
                             help='The OCSP response to lint'
                             )

    args = parser.parse_args(cli_args)

    doc_validator = ocsp.create_pkix_ocsp_response_validator_container(
        [
            ocsp.create_response_decoder(),
            pkix.create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
            pkix.create_extension_decoder(extension.EXTENSION_MAPPINGS),
        ],
        []
    )

    if args.command == 'validations':
        print(report.report_included_validations(doc_validator))

        return 0
    else:
        ocsp_response = loader.load_ocsp_response(args.file, args.file.name)

        results = doc_validator.validate(ocsp_response.root)

        print(args.format(results, args.severity))

        return report.get_findings_count(results, args.severity)


if __name__ == "__main__":
    sys.exit(main())



from typing import List, Set

import validators
from pyasn1.type.univ import ObjectIdentifier
from pyasn1_alt_modules import rfc5280, rfc2985

from pkilint import document, validation, oid
from pkilint.itu import x520_name

ATTRIBUTE_TYPE_MAPPINGS = {
    **x520_name.ATTRIBUTE_TYPE_MAPPINGS,
    **rfc2985._certificateAttributesMapUpdate,
    **rfc5280.certificateAttributesMap,
}


def get_name_attributes_by_type(name_node, type_oid):
    atvs = []

    for rdn_idx, rdn in name_node.children['rdnSequence'].children.items():
        for atv_idx, atv in rdn.children.items():
            if atv.children['type'].pdu == type_oid:
                atvs.append((atv, (int(rdn_idx), int(atv_idx))))

    return atvs


class EmptyNameValidator(validation.Validator):
    VALIDATION_NAME_IS_EMPTY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.name_empty'
    )

    def __init__(self, **kwargs):
        super().__init__(validations=[self.VALIDATION_NAME_IS_EMPTY],
                         pdu_class=rfc5280.RDNSequence, **kwargs)

    def validate(self, node):
        if len(node.children) == 0:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NAME_IS_EMPTY,
            )


class RDNContainsUniqueTypesValidator(validation.Validator):
    VALIDATION_ATTRIBUTE_TYPES_NOT_UNIQUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.rdn_contains_duplicate_attribute_types'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_ATTRIBUTE_TYPES_NOT_UNIQUE,
                         pdu_class=rfc5280.RelativeDistinguishedName
                         )

    def validate(self, node):
        oids = set()
        for child in node.children.values():
            oid = child.children['type'].pdu
            if oid in oids:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_ATTRIBUTE_TYPES_NOT_UNIQUE,
                    f'Multiple attributes of type "{str(oid)}"'
                )

            oids.add(oid)


class NameAttributeTypesValidatorBase(validation.Validator):
    def __init__(self, *, expected_oid_set: Set[ObjectIdentifier],
                 validation: validation.ValidationFinding, **kwargs
                 ):
        self.expected_oid_set = expected_oid_set
        self.validation = validation

        super().__init__(
            validations=[self.validation],
            pdu_class=rfc5280.RDNSequence,
            **kwargs
        )

    def validate_attributes(self, node: document.PDUNode,
                            attributes: List[ObjectIdentifier]
                            ) -> validation.ValidationResult:
        pass

    def validate(self, node):
        attributes = []

        for rdn in node.children.values():
            attributes += [
                atv.children['type'].pdu
                for atv in rdn.children.values()
            ]

        return self.validate_attributes(node, attributes)


class PermittedAttributeTypeValidator(NameAttributeTypesValidatorBase):
    def __init__(self, *, allowed_oid_set, validation):
        super().__init__(
            expected_oid_set=allowed_oid_set,
            validation=validation
        )

    def validate_attributes(self, node, attributes):
        attributes = set(attributes)

        prohibited_oids = attributes - self.expected_oid_set

        if len(prohibited_oids) > 0:
            oids_str = oid.format_oids(prohibited_oids)

            raise validation.ValidationFindingEncountered(
                self.validation,
                f'Prohibited attribute types: {oids_str}'
            )


class RequiredAttributeTypeValidator(NameAttributeTypesValidatorBase):
    def __init__(self, *, required_oid_set, validation):
        super().__init__(
            expected_oid_set=required_oid_set,
            validation=validation
        )

    def validate_attributes(self, node, attributes):
        attributes = set(attributes)

        missing_oids = self.expected_oid_set - attributes

        if len(missing_oids) > 0:
            oids_str = oid.format_oids(missing_oids)

            raise validation.ValidationFindingEncountered(
                self.validation,
                f'Required attribute types not present: {oids_str}'
            )


class NameDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(decode_func=decode_func,
                         pdu_class=rfc5280.AttributeTypeAndValue,
                         **kwargs
                         )


class IssuerSubjectNameBinaryEqualValidator(validation.DEREqualityValidator):
    VALIDATION_ISSUER_AND_SUBJECT_NOT_DER_EQUAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'pkix.issuer_and_subject_dn_not_binary_equal'
    )

    def _retrieve_subject_issuer_dn(self, node):
        return node.navigate(self._subject_document_issuer_dn_path)

    def __init__(self, *, subject_document_issuer_dn_path, **kwargs):
        self._subject_document_issuer_dn_path = subject_document_issuer_dn_path

        super().__init__(other_node_retriever=self._retrieve_subject_issuer_dn,
                         validation=self.VALIDATION_ISSUER_AND_SUBJECT_NOT_DER_EQUAL,
                         **kwargs
                         )


class DomainComponentValidDomainNameValidator(validation.Validator):
    VALIDATION_NAME_DC_NOT_A_VALID_DOMAIN_NAME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.name_domain_components_invalid_domain_name'
    )

    def __init__(self, **kwargs):
        super().__init__(
            validations=[self.VALIDATION_NAME_DC_NOT_A_VALID_DOMAIN_NAME],
            **kwargs
        )

    def validate(self, node):
        atvs = node.document.get_name_attributes_by_type(
            rfc5280.id_domainComponent, f':{node.path}'
        )

        try:
            components = [
                str(a.navigate('value.domainComponent').pdu)
                for a, _ in atvs
            ]
        except document.PDUNavigationFailedError:
            return

        components.reverse()

        domain_name = '.'.join(components)

        if len(domain_name) > 0:
            ret = validators.domain(domain_name)

            if not isinstance(ret, bool) or not ret:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_NAME_DC_NOT_A_VALID_DOMAIN_NAME,
                    f'Invalid domain name in domainComponents: {domain_name}'
                )


class DuplicateAttributeTypeValidator(NameAttributeTypesValidatorBase):
    def __init__(
            self, *, allowed_duplicate_oid_set, validation
    ):
        super().__init__(
            expected_oid_set=allowed_duplicate_oid_set,
            validation=validation,
        )

    def validate_attributes(self, node, attributes):
        duplicates = set((
            o for o in attributes if attributes.count(o) > 1
        ))

        disallowed_duplicates = duplicates - self.expected_oid_set

        if len(disallowed_duplicates) > 0:
            oids_str = oid.format_oids(disallowed_duplicates)

            raise validation.ValidationFindingEncountered(
                self.validation,
                f'Prohibited multiple instances of attribute types: {oids_str}'
            )



import validators
from pyasn1.type import char
from pyasn1_alt_modules import rfc5280, rfc8398

from pkilint import validation, pkix, oid
from pkilint.cabf import cabf_name, cabf_constants
from pkilint.cabf.cabf_name import OrganizationIdentifierAttributeValidator
from pkilint.cabf.smime.smime_constants import Generation, ValidationLevel
from pkilint.iso import lei
from pkilint.itu import x520_name
from pkilint.pkix import certificate, name

SHALL = pkix.Rfc2119Word.SHALL
SHALL_NOT = pkix.Rfc2119Word.SHALL_NOT
MAY = pkix.Rfc2119Word.MAY

_MV_ATTRIBUTES = {
    rfc5280.id_at_commonName: (MAY, MAY, MAY),
    rfc5280.id_at_organizationName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_organizationalUnitName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_organizationIdentifier: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_givenName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_surname: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_pseudonym: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_serialNumber: (MAY, MAY, MAY),
    rfc5280.id_emailAddress: (MAY, MAY, MAY),
    rfc5280.id_at_title: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_streetAddress: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_localityName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_stateOrProvinceName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_postalCode: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_countryName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
}

_OV_ATTRIBUTES = {
    rfc5280.id_at_commonName: (MAY, MAY, MAY),
    rfc5280.id_at_organizationName: (SHALL, SHALL, SHALL),
    rfc5280.id_at_organizationalUnitName: (MAY, MAY, MAY),
    x520_name.id_at_organizationIdentifier: (SHALL, SHALL, SHALL),
    rfc5280.id_at_givenName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_surname: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_pseudonym: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_serialNumber: (MAY, MAY, MAY),
    rfc5280.id_emailAddress: (MAY, MAY, MAY),
    rfc5280.id_at_title: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_streetAddress: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_localityName: (MAY, MAY, MAY),
    rfc5280.id_at_stateOrProvinceName: (MAY, MAY, MAY),
    x520_name.id_at_postalCode: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_countryName: (MAY, MAY, MAY),
}

_SV_ATTRIBUTES = {
    rfc5280.id_at_commonName: (MAY, MAY, MAY),
    rfc5280.id_at_organizationName: (SHALL, SHALL, SHALL),
    rfc5280.id_at_organizationalUnitName: (MAY, MAY, MAY),
    x520_name.id_at_organizationIdentifier: (SHALL, SHALL, SHALL),
    rfc5280.id_at_givenName: (MAY, MAY, MAY),
    rfc5280.id_at_surname: (MAY, MAY, MAY),
    rfc5280.id_at_pseudonym: (MAY, MAY, MAY),
    rfc5280.id_at_serialNumber: (MAY, MAY, MAY),
    rfc5280.id_emailAddress: (MAY, MAY, MAY),
    rfc5280.id_at_title: (MAY, MAY, MAY),
    x520_name.id_at_streetAddress: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_localityName: (MAY, MAY, MAY),
    rfc5280.id_at_stateOrProvinceName: (MAY, MAY, MAY),
    x520_name.id_at_postalCode: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_countryName: (MAY, MAY, MAY),
}

_IV_ATTRIBUTES = {
    rfc5280.id_at_commonName: (MAY, MAY, MAY),
    rfc5280.id_at_organizationName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_organizationalUnitName: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    x520_name.id_at_organizationIdentifier: (SHALL_NOT, SHALL_NOT, SHALL_NOT),
    rfc5280.id_at_givenName: (MAY, MAY, MAY),
    rfc5280.id_at_surname: (MAY, MAY, MAY),
    rfc5280.id_at_pseudonym: (MAY, MAY, MAY),
    rfc5280.id_at_serialNumber: (MAY, MAY, MAY),
    rfc5280.id_emailAddress: (MAY, MAY, MAY),
    rfc5280.id_at_title: (MAY, MAY, MAY),
    x520_name.id_at_streetAddress: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_localityName: (MAY, MAY, MAY),
    rfc5280.id_at_stateOrProvinceName: (MAY, MAY, MAY),
    x520_name.id_at_postalCode: (MAY, MAY, SHALL_NOT),
    rfc5280.id_at_countryName: (MAY, MAY, MAY),
}

_GENERATION_INDEXES = {
    Generation.LEGACY: 0,
    Generation.MULTIPURPOSE: 1,
    Generation.STRICT: 2,
}

_VALIDATION_LEVEL_TO_TABLE = {
    ValidationLevel.MAILBOX: _MV_ATTRIBUTES,
    ValidationLevel.ORGANIZATION: _OV_ATTRIBUTES,
    ValidationLevel.SPONSORED: _SV_ATTRIBUTES,
    ValidationLevel.INDIVIDUAL: _IV_ATTRIBUTES,
}

_VALIDATION_LEVEL_TO_OTHER_ATTRIBUTE_ALLOWANCE = {
    ValidationLevel.MAILBOX: (False, False, False),
    ValidationLevel.ORGANIZATION: (True, False, False),
    ValidationLevel.SPONSORED: (True, False, False),
    ValidationLevel.INDIVIDUAL: (True, False, False),
}

_REQUIRED_ONE_OF_N = {
    (ValidationLevel.SPONSORED, Generation.LEGACY): {rfc5280.id_at_givenName, rfc5280.id_at_surname,
                                                     rfc5280.id_at_pseudonym, rfc5280.id_at_commonName},
    (ValidationLevel.INDIVIDUAL, Generation.LEGACY): {rfc5280.id_at_givenName, rfc5280.id_at_surname,
                                                      rfc5280.id_at_pseudonym, rfc5280.id_at_commonName},
    (ValidationLevel.SPONSORED, Generation.MULTIPURPOSE): {rfc5280.id_at_givenName, rfc5280.id_at_surname,
                                                           rfc5280.id_at_pseudonym},
    (ValidationLevel.INDIVIDUAL, Generation.MULTIPURPOSE): {rfc5280.id_at_givenName, rfc5280.id_at_surname,
                                                            rfc5280.id_at_pseudonym},
    (ValidationLevel.SPONSORED, Generation.STRICT): {rfc5280.id_at_givenName, rfc5280.id_at_surname,
                                                     rfc5280.id_at_pseudonym},
    (ValidationLevel.INDIVIDUAL, Generation.STRICT): {rfc5280.id_at_givenName, rfc5280.id_at_surname,
                                                      rfc5280.id_at_pseudonym},
}


class SubscriberSubjectValidator(validation.Validator):
    VALIDATION_MISSING_ATTRIBUTE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.missing_required_attribute'
    )

    VALIDATION_PROHIBITED_ATTRIBUTE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.prohibited_attribute'
    )

    VALIDATION_MIXED_NAME_AND_PSEUDONYM_ATTRIBUTES = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.mixed_name_and_pseudonym_attributes'
    )

    def __init__(self, validation_level, generation):
        super().__init__(validations=[
            self.VALIDATION_PROHIBITED_ATTRIBUTE,
            self.VALIDATION_MISSING_ATTRIBUTE,
            self.VALIDATION_MIXED_NAME_AND_PSEUDONYM_ATTRIBUTES,
        ],
            pdu_class=rfc5280.RDNSequence,
            predicate=lambda n: n.path != 'certificate.tbsCertificate.issuer.rdnSequence')

        self._attribute_table = {
            k: v[_GENERATION_INDEXES[generation]] for k, v in _VALIDATION_LEVEL_TO_TABLE[validation_level].items()
        }

        self._required_attributes = {k for k, v in self._attribute_table.items() if v == SHALL}
        self._prohibited_attributes = {k for k, v, in self._attribute_table.items() if v == SHALL_NOT}
        self._required_one_of_n_attributes = _REQUIRED_ONE_OF_N.get((validation_level, generation))

        self._allow_other_oids = _VALIDATION_LEVEL_TO_OTHER_ATTRIBUTE_ALLOWANCE[validation_level][
            _GENERATION_INDEXES[generation]]

    def validate(self, node):
        findings = []

        attributes = set()
        for rdn in node.children.values():
            attributes.update((atv.children['type'].pdu for atv in rdn.children.values()))

        findings.extend((
            validation.ValidationFindingDescription(self.VALIDATION_MISSING_ATTRIBUTE,
                                                    f'Missing required attribute: {a}')
            for a in self._required_attributes - attributes
        ))

        if self._required_one_of_n_attributes and len(self._required_one_of_n_attributes.intersection(attributes)) == 0:
            oids = oid.format_oids(self._required_one_of_n_attributes)

            findings.append(validation.ValidationFindingDescription(self.VALIDATION_MISSING_ATTRIBUTE,
                                                                    f'Missing one of these required attributes: {oids}'))

        findings.extend((
            validation.ValidationFindingDescription(self.VALIDATION_PROHIBITED_ATTRIBUTE,
                                                    f'Prohibited attribute: {a}')
            for a in self._prohibited_attributes.intersection(attributes)
        ))

        if rfc5280.id_at_pseudonym in attributes and (
                any({rfc5280.id_at_givenName, rfc5280.id_at_surname}.intersection(attributes))):
            findings.append(
                validation.ValidationFindingDescription(self.VALIDATION_MIXED_NAME_AND_PSEUDONYM_ATTRIBUTES, None))

        if not self._allow_other_oids:
            findings.extend((
                validation.ValidationFindingDescription(self.VALIDATION_PROHIBITED_ATTRIBUTE,
                                                        f'Prohibited other attribute: {a}')
                for a in attributes - set(self._attribute_table.keys())
            ))

        return validation.ValidationResult(self, node, findings)


def create_subscriber_certificate_subject_validator_container(
        validation_level, generation
):
    dn_validators = [
        SubscriberSubjectValidator(validation_level, generation),
        SubjectAlternativeNameContainsSubjectEmailAddressesValidator(),
        cabf_name.ValidCountryValidator(),
        CommonNameValidator(validation_level, generation),
        OrganizationIdentifierAttributeValidator(relax_stateprovince_syntax=True, additional_schemes={
            'LEI': cabf_constants.RegistrationSchemeNamingConvention(
                cabf_constants.RegistrationSchemeCountryIdentifierType.XG,
                False, True
            ),
            'GOV': cabf_constants.RegistrationSchemeNamingConvention(
                cabf_constants.RegistrationSchemeCountryIdentifierType.ISO3166,
                True, False
            ),
            'INT': cabf_constants.RegistrationSchemeNamingConvention(
                cabf_constants.RegistrationSchemeCountryIdentifierType.XG,
                False, False
            )
        }),
        OrganizationIdentifierLeiValidator(),
        OrganizationIdentifierCountryNameConsistentValidator(),
        cabf_name.RelativeDistinguishedNameContainsOneElementValidator(),
    ]

    return certificate.create_subject_validator_container(
        dn_validators, pdu_class=rfc5280.Name,
        predicate=lambda n: n.path != 'certificate.tbsCertificate.issuer'
    )


class SubjectAlternativeNameContainsSubjectEmailAddressesValidator(
    validation.Validator
):
    VALIDATION_EMAIL_ADDRESS_IN_CN_MISSING_FROM_SAN = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.email_address_in_common_name_not_in_san'
    )

    VALIDATION_EMAIL_ADDRESS_IN_ATTRIBUTE_MISSING_FROM_SAN = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.smime.email_address_in_attribute_not_in_san'
    )

    VALIDATION_UNPARSED_ATTRIBUTE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'cabf.smime.unparsed_attribute_value_encountered'
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_EMAIL_ADDRESS_IN_CN_MISSING_FROM_SAN,
                self.VALIDATION_EMAIL_ADDRESS_IN_ATTRIBUTE_MISSING_FROM_SAN,
                self.VALIDATION_UNPARSED_ATTRIBUTE,
            ],
            pdu_class=rfc5280.AttributeTypeAndValue,
            # emailAddress presence in SAN is checked by PKIX lint
            predicate=lambda n: n.children['type'].pdu != rfc5280.id_emailAddress
        )

    def validate(self, node):
        oid = node.children['type'].pdu
        value = node.children['value']

        while True:
            if len(value.children) != 1:
                value = None
                break
            else:
                _, value = value.child

                if len(value.children) == 0:
                    if isinstance(value.pdu, char.AbstractCharacterString):
                        break

        if value is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNPARSED_ATTRIBUTE,
                f'Unparsed attribute {str(oid)} encountered'
            )

        value_str = str(value.pdu)

        if bool(validators.email(value_str)):
            san_ext = node.document.get_extension_by_oid(
                rfc5280.id_ce_subjectAltName
            )
            if san_ext is None:
                email_sans = set()
            else:
                email_sans = set((
                    str(gn.child[1].pdu)
                    for gn in san_ext[0].navigate('extnValue.subjectAltName').children.values()
                    if gn.child[0] == 'rfc822Name'
                ))

            if value_str not in email_sans:
                if oid == rfc5280.id_at_commonName:
                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_EMAIL_ADDRESS_IN_CN_MISSING_FROM_SAN,
                        value_str
                    )
                else:
                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_EMAIL_ADDRESS_IN_ATTRIBUTE_MISSING_FROM_SAN,
                        f'Attribute {str(oid)} with value "{value_str}" not found in SAN'
                    )


class CommonNameValidator(validation.Validator):
    VALIDATION_COMMON_NAME_UNKNOWN_VALUE_SOURCE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.common_name_value_unknown_source'
    )

    VALIDATION_UNPARSED_COMMON_NAME_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'cabf.smime.unparsed_common_name_value'
    )

    def __init__(self, validation_level, generation):
        super().__init__(
            validations=[self.VALIDATION_COMMON_NAME_UNKNOWN_VALUE_SOURCE, self.VALIDATION_UNPARSED_COMMON_NAME_VALUE],
            pdu_class=rfc5280.X520CommonName
        )

        self._validation_level = validation_level
        self._generation = generation

    @staticmethod
    def _is_value_in_dirstring_atvs(atvs, expected_value_node):
        for atv in atvs:
            try:
                # get the value contained within the DirectoryString-encoded ATV value
                _, atv_dirstring_value_node = atv.children['value'].child
                _, value = atv_dirstring_value_node.child
            except ValueError:
                # skip unparsed field

                continue

            if str(value.pdu) == str(expected_value_node.pdu):
                return True

        return False

    def validate(self, node):
        try:
            _, cn_value_node = node.child
        except ValueError:
            raise validation.ValidationFindingEncountered(self.VALIDATION_UNPARSED_COMMON_NAME_VALUE)

        parent_name_node = next((n for n in node.parents if isinstance(n.pdu, rfc5280.Name)))

        if self._validation_level in {ValidationLevel.SPONSORED, ValidationLevel.INDIVIDUAL}:
            # legacy sponsored and individual profiles allow the Personal Name in CN without being in other
            # subject attributes
            if self._generation == Generation.LEGACY:
                return

            # we don't need the index
            pseudonym_nodes = [t[0] for t in
                               name.get_name_attributes_by_type(parent_name_node, rfc5280.id_at_pseudonym)]

            if CommonNameValidator._is_value_in_dirstring_atvs(pseudonym_nodes, cn_value_node):
                return

            # if there's a GN or SN, assume it's in the CN
            if (
                    any(name.get_name_attributes_by_type(parent_name_node, rfc5280.id_at_givenName)) or
                    any(name.get_name_attributes_by_type(parent_name_node, rfc5280.id_at_surname))):
                return
        elif self._validation_level == ValidationLevel.ORGANIZATION:
            orgname_nodes = [t[0] for t in
                             name.get_name_attributes_by_type(parent_name_node, rfc5280.id_at_organizationName)]

            if CommonNameValidator._is_value_in_dirstring_atvs(orgname_nodes, cn_value_node):
                return

        email_addresses = get_email_addresses_from_san(node.document)

        if str(cn_value_node.pdu) not in email_addresses:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_COMMON_NAME_UNKNOWN_VALUE_SOURCE,
                f'Unknown CN value source: "{str(cn_value_node.pdu)}"'
            )


class OrganizationIdentifierLeiValidator(validation.Validator):
    VALIDATION_INVALID_ORGID_LEI_FORMAT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, 'cabf.smime.invalid_lei_scheme_format'
    )

    _LEI_PREFIX = 'LEIXG-'

    def __init__(self):
        super().__init__(validations=[
            lei.VALIDATION_INVALID_LEI_CHECKSUM, lei.VALIDATION_INVALID_LEI_FORMAT,
            self.VALIDATION_INVALID_ORGID_LEI_FORMAT
        ],
            pdu_class=x520_name.X520OrganizationIdentifier,
            predicate=lambda n: any(n.children) and str(n.child[1].pdu).startswith('LEI')
        )

    def validate(self, node):
        value = str(node.child[1].pdu)

        if not value.startswith(self._LEI_PREFIX):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_ORGID_LEI_FORMAT,
                f'Invalid Organization Identifier format: {value}'
            )

        lei_value = value[len(self._LEI_PREFIX):]

        lei.validate_lei(lei_value)


class OrganizationIdentifierCountryNameConsistentValidator(validation.Validator):
    VALIDATION_ORGID_COUNTRYNAME_INCONSISTENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.org_identifier_and_country_name_attribute_inconsistent'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_ORGID_COUNTRYNAME_INCONSISTENT,
                         pdu_class=rfc5280.X520countryName)

    def validate(self, node):
        country_name_value = str(node.pdu)

        for atv, _ in node.document.get_subject_attributes_by_type(x520_name.id_at_organizationIdentifier):
            attr_value_node = atv.navigate('value')

            try:
                _, x520_dirstring_value_node = attr_value_node.child
            except ValueError:
                continue

            _, x520_value_node = x520_dirstring_value_node.child

            x520_value_str = str(x520_value_node.pdu)

            m = cabf_name.ORG_ID_REGEX.match(x520_value_str)

            if m is None:
                continue

            orgid_country_name = m['country']

            # skip this orgId attribute if it doesn't contain a countryName or contains XG
            if not orgid_country_name or orgid_country_name.upper() == 'XG':
                continue

            if orgid_country_name != country_name_value:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_ORGID_COUNTRYNAME_INCONSISTENT,
                    f'CountryName attribute value: "{country_name_value}", '
                    f'OrganizationIdentifier attribute country name value: "{orgid_country_name}"'
                )


def get_email_addresses_from_san(cert_document):
    san_ext_and_idx = cert_document.get_extension_by_oid(rfc5280.id_ce_subjectAltName)

    if san_ext_and_idx is None:
        return []

    san_ext, _ = san_ext_and_idx

    email_addresses = []
    for gn in san_ext.navigate('extnValue.subjectAltName').children.values():
        name, value = gn.child

        if name == 'rfc822Name':
            email_addresses.append(value.pdu)
        elif name == 'otherName' and value.navigate('type-id').pdu == rfc8398.id_on_SmtpUTF8Mailbox:
            email_addresses.append(value.navigate('value').pdu)

    return email_addresses



import validators

from pkilint import validation
from pkilint.msft import asn1


class UserPrincipalNameSyntaxValidator(validation.Validator):
    VALIDATION_INVALID_UPN_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'msft.invalid_user_principal_name_syntax'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_INVALID_UPN_SYNTAX], pdu_class=asn1.UserPrincipalName)

    def validate(self, node):
        value = str(node.pdu)

        if not validators.email(value):
            raise validation.ValidationFindingEncountered(self.VALIDATION_INVALID_UPN_SYNTAX,
                                                          f'Invalid UPN syntax: "{value}"')



import base64
import binascii
import re

from cryptography.hazmat.primitives import hashes
from pyasn1_alt_modules import rfc5280

from pkilint import validation, document
from pkilint.cabf.asn1 import ev_guidelines
from pkilint.cabf.cabf_name import ValidCountryCodeValidatorBase, ORG_ID_REGEX
from pkilint.cabf.serverauth import serverauth_constants
from pkilint.itu import x520_name
from pkilint.pkix import name, general_name


class ValidJurisdictionCountryValidator(ValidCountryCodeValidatorBase):
    """Validates that the jurisdictionCountryName value conforms to EVG 9.2.4."""

    VALIDATION_INVALID_COUNTRY_CODE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.invalid_jurisdiction_country_code'
    )

    def __init__(self):
        super().__init__(
            type_oid=ev_guidelines.id_evat_jurisdiction_countryName,
            value_path='value.eVGJurisdictionCountryName',
            checked_validation=self.VALIDATION_INVALID_COUNTRY_CODE
        )


class ValidBusinessCategoryValidator(validation.Validator):
    """Validates that the businessCategory value conforms to EVG 9.2.3."""

    VALIDATION_INVALID_BUSINESS_CATEGORY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.ev_guidelines.invalid_business_category'
    )

    _ALLOWED_VALUES = {
        'Private Organization',
        'Government Entity',
        'Business Entity',
        'Non-Commercial Entity'
    }

    def __init__(self):
        super().__init__(
            pdu_class=x520_name.X520BusinessCategory,
            validations=[self.VALIDATION_INVALID_BUSINESS_CATEGORY]
        )

    def validate(self, node):
        # BusinessCategory is a CHOICE so retrieve the child node value
        business_category = str(node.child[1].pdu)

        if business_category not in self._ALLOWED_VALUES:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_BUSINESS_CATEGORY,
                f'Invalid business category: {business_category}'
            )


class OrganizationIdentifierConsistentSubjectAndExtensionValidator(validation.Validator):
    """Validates that the content of the organizationIdentifier subject attributes and the organizationIdentifier
    extension are consistent, as per EVG 9.2.8 and 9.2.9."""

    VALIDATION_CABF_ORG_ID_NO_EXT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.organization_identifier_extension_absent'
    )

    VALIDATION_CABF_ORG_ID_INVALID_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.FATAL,
        'cabf.serverauth.organization_identifier_invalid_syntax'
    )

    VALIDATION_CABF_ORG_ID_MISMATCHED_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.organization_identifier_mismatched_scheme'
    )

    VALIDATION_CABF_ORG_ID_MISMATCHED_COUNTRY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.organization_identifier_mismatched_country_code'
    )

    VALIDATION_CABF_ORG_ID_MISMATCHED_SP = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.organization_identifier_mismatched_state_province'
    )

    VALIDATION_CABF_ORG_ID_MISMATCHED_REFERENCE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.organization_identifier_mismatched_registration_reference'
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_CABF_ORG_ID_NO_EXT,
                self.VALIDATION_CABF_ORG_ID_INVALID_SYNTAX,
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_SCHEME,
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_COUNTRY,
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_SP,
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_REFERENCE,
            ],
            pdu_class=x520_name.X520OrganizationIdentifier,
            predicate=lambda n: any(n.children)
        )

    def validate(self, node):
        ext_and_idx = node.document.get_extension_by_oid(
            ev_guidelines.id_CABFOrganizationIdentifier
        )

        if ext_and_idx is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CABF_ORG_ID_NO_EXT
            )

        ext, _ = ext_and_idx
        try:
            ext = ext.navigate('extnValue.cABFOrganizationIdentifier')
        except document.PDUNavigationFailedError:
            return

        attr_value = str(node.child[1].pdu)

        m = ORG_ID_REGEX.match(attr_value)
        if m is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CABF_ORG_ID_INVALID_SYNTAX,
                f'Invalid syntax: {attr_value}'
            )

        findings = []
        ext_scheme = str(ext.children['registrationSchemeIdentifier'].pdu)
        if m['scheme'] != ext_scheme:
            findings.append(validation.ValidationFindingDescription(
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_SCHEME,
                f'Mismatched scheme: subject: "{m["scheme"]}", extension: '
                f'"{ext_scheme}"'
            ))

        ext_country = str(ext.children['registrationCountry'].pdu)
        if m['country'] != ext_country:
            findings.append(validation.ValidationFindingDescription(
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_COUNTRY,
                f'Mismatched country: subject: "{m["country"]}", extension: '
                f'"{ext_country}"'
            ))

        ext_sp_node = ext.children.get('registrationStateOrProvince')
        ext_sp = None if ext_sp_node is None else str(ext_sp_node.pdu)
        if m['sp'] is None and ext_sp is not None:
            findings.append(validation.ValidationFindingDescription(
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_SP,
                'Extension has state/province value but subject does not'
            ))
        elif m['sp'] is not None and ext_sp is None:
            findings.append(validation.ValidationFindingDescription(
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_SP,
                'Extension does not have state/province value but subject does'
            ))
        elif m['sp'] is not None and ext_sp is not None:
            if ext_sp != m['sp']:
                findings.append(validation.ValidationFindingDescription(
                    self.VALIDATION_CABF_ORG_ID_MISMATCHED_SP,
                    f'Mismatched state/province value: subject: "{m["sp"]}", '
                    f'extension: "{ext_sp}"'
                ))

        ext_reg_ref = str(ext.children['registrationReference'].pdu)
        if m['reference'] != ext_reg_ref:
            findings.append(validation.ValidationFindingDescription(
                self.VALIDATION_CABF_ORG_ID_MISMATCHED_REFERENCE,
                f'Mismatched Registration Reference: subject: "{m["reference"]}"'
                f', extension: "{ext_reg_ref}"'
            ))

        return validation.ValidationResult(self, node, findings)


class X520NameAttributeValueLengthValidator(validation.Validator):
    """Validates that the length of X520Name values does not exceed the limit stated in BR 7.1.4.2."""

    VALIDATION_NAME_VALUE_TOO_LONG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.name_attribute_value_too_long'
    )

    _MAX_NAME_LENGTH_CHARS = 64

    def __init__(self):
        super().__init__(validations=self.VALIDATION_NAME_VALUE_TOO_LONG, pdu_class=rfc5280.X520name)

    def validate(self, node):
        _, value_node = node.child

        value = str(value_node.pdu)

        if len(value) > self._MAX_NAME_LENGTH_CHARS:
            raise validation.ValidationFindingEncountered(
                self._validations[0],
                f'Attribute value exceeds maximum length of {self._MAX_NAME_LENGTH_CHARS}: "{value}"'
            )


class DomainComponentAttributeValueLengthValidator(validation.Validator):
    """Validates that the length of domainComponent values does not exceed the limit stated in BR 7.1.4.2."""

    VALIDATION_DC_ATTRIBUTE_VALUE_TOO_LONG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.domain_component_attribute_value_length_too_long'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_DC_ATTRIBUTE_VALUE_TOO_LONG,
                         pdu_class=rfc5280.DomainComponent)

    def validate(self, node):
        value_str = str(node.pdu)

        if len(value_str) > 63:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_DC_ATTRIBUTE_VALUE_TOO_LONG
            )


class AttributeValueDirectoryStringValidator(validation.Validator):
    """Validates that DirectoryString attributes are encoded as per 7.1.4.2."""

    VALIDATION_ATTRIBUTE_VALUE_INVALID_ENCODING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.attribute_value_invalid_encoding_type'
    )

    _ENFORCED_ATTRIBUTE_TYPE_OIDS = {
        rfc5280.id_at_stateOrProvinceName,
        rfc5280.id_at_localityName,
        x520_name.id_at_postalCode,
        x520_name.id_at_streetAddress,
        rfc5280.id_at_organizationName,
        rfc5280.id_at_surname,
        rfc5280.id_at_givenName,
        rfc5280.id_at_organizationalUnitName,
        rfc5280.id_at_commonName,
        x520_name.id_at_businessCategory,
        ev_guidelines.id_evat_jurisdiction_stateOrProvinceName,
        ev_guidelines.id_evat_jurisdiction_localityName,
        x520_name.id_at_organizationIdentifier,
    }

    def __init__(self):
        super().__init__(validations=self.VALIDATION_ATTRIBUTE_VALUE_INVALID_ENCODING,
                         pdu_class=rfc5280.AttributeTypeAndValue,
                         predicate=lambda n: n.navigate('type').pdu in self._ENFORCED_ATTRIBUTE_TYPE_OIDS)

    def validate(self, node):
        value_node = node.navigate('value')

        if not any(value_node.children):
            # unparsed attribute
            return

        _, parsed_attr = value_node.child
        directory_string_choice_name, _ = parsed_attr.child

        if directory_string_choice_name not in {'utf8String', 'printableString'}:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ATTRIBUTE_VALUE_INVALID_ENCODING,
                f'Invalid attribute value encoding: {directory_string_choice_name}'
            )


class AttributeOrderEncodingValidator(validation.Validator):
    """Validates that the encoded order of subject attributes conforms to the list in BR 7.1.4.2."""

    VALIDATION_INVALID_RDN_ORDER = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.invalid_rdn_order'
    )

    _ENFORCED_ATTRIBUTE_TYPE_OIDS = [
        rfc5280.id_domainComponent,
        rfc5280.id_at_countryName,
        rfc5280.id_at_stateOrProvinceName,
        rfc5280.id_at_localityName,
        x520_name.id_at_postalCode,
        x520_name.id_at_streetAddress,
        rfc5280.id_at_organizationName,
        rfc5280.id_at_surname,
        rfc5280.id_at_givenName,
        rfc5280.id_at_organizationalUnitName,
        rfc5280.id_at_commonName,
    ]

    def __init__(self):
        super().__init__(validations=self.VALIDATION_INVALID_RDN_ORDER,
                         path='certificate.tbsCertificate.subject.rdnSequence')

    def validate(self, node):
        enforced_order_cursor_idx = None

        # assume each RDN has one ATV
        for rdn in node.children.values():
            attr_oid = rdn.children['0'].children['type'].pdu

            try:
                idx = self._ENFORCED_ATTRIBUTE_TYPE_OIDS.index(attr_oid)
            except ValueError:
                continue

            if enforced_order_cursor_idx is None:
                enforced_order_cursor_idx = idx
            elif idx < enforced_order_cursor_idx:
                current_rdn_oid = str(attr_oid)
                enforced_order_oid = str(self._ENFORCED_ATTRIBUTE_TYPE_OIDS[enforced_order_cursor_idx])

                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_INVALID_RDN_ORDER,
                    f'Invalid RDN order: {current_rdn_oid} follows {enforced_order_oid}')
            else:
                enforced_order_cursor_idx = idx


class ServerauthRelativeDistinguishedNameContainsOneElementValidator(validation.Validator):
    """Validates that each RelativeDistguishedName contains exactly one AttributeTypeAndValue, as per BR 7.1.4.2."""

    VALIDATION_RDN_CONTAINS_MULTIPLE_ATVS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.rdn_contains_multiple_atvs'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_RDN_CONTAINS_MULTIPLE_ATVS],
                         pdu_class=rfc5280.RelativeDistinguishedName)

    def validate(self, node):
        child_count = len(node.children)

        if child_count > 1:
            raise validation.ValidationFindingEncountered(self.VALIDATION_RDN_CONTAINS_MULTIPLE_ATVS)


class ServerauthDuplicateAttributeTypeValidator(name.DuplicateAttributeTypeValidator):
    """Validates that only specified attributes may appear multiple times as per BR 7.1.2.3 and 7.1.2.7.4."""

    VALIDATION_DUPLICATE_ATTRIBUTE_TYPES = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.prohibited_duplicate_attribute_type'
    )

    def __init__(self, certificate_type):
        allowed_oids = set()

        if certificate_type in {serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE,
                                serverauth_constants.CertificateType.OV_PRE_CERTIFICATE}:
            allowed_oids.add(rfc5280.id_domainComponent)

        if certificate_type not in {serverauth_constants.CertificateType.DV_FINAL_CERTIFICATE,
                                    serverauth_constants.CertificateType.DV_PRE_CERTIFICATE}:
            allowed_oids.add(x520_name.id_at_streetAddress)

        super().__init__(
            allowed_duplicate_oid_set=allowed_oids,
            validation=self.VALIDATION_DUPLICATE_ATTRIBUTE_TYPES
        )


# BR 7.1.2.7.12
class DnsNameLdhLabelSyntaxValidator(validation.Validator):
    """Validates that each dNSName conforms to the syntax in BR 7.1.2.7.12."""

    VALIDATION_INVALID_DNSNAME_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.invalid_dnsname_syntax'
    )

    VALIDATION_PROHIBITED_RESERVED_LABEL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.dnsname_contains_prohibited_reserved_label'
    )

    VALIDATION_INVALID_PLABEL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.invalid_plabel_encoding'
    )

    _ACE_REGEX = re.compile(r'^(?P<tag>.{2})--(?P<ace>.+)$', re.RegexFlag.IGNORECASE)
    _NON_LDH_CHAR_REGEX = re.compile(r'[^a-zA-Z0-9-]')

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_INVALID_DNSNAME_SYNTAX,
                self.VALIDATION_PROHIBITED_RESERVED_LABEL,
                self.VALIDATION_INVALID_PLABEL],
            predicate=general_name.create_generalname_type_predicate('dNSName')
         )

    def _is_valid_ldh_syntax(self, label):
        label_len = len(label)

        return (0 < label_len <= 63 and
                self._NON_LDH_CHAR_REGEX.search(label) is None and
                not label.startswith('-') and not label.endswith('-')
                )

    def validate(self, node):
        value = str(node.pdu)

        in_nc = any((isinstance(n.pdu, rfc5280.NameConstraints) for n in node.parents))

        if in_nc and len(value) == 0:
            return

        fqdn = value[2:] if not in_nc and value.startswith('*.') else value

        labels = fqdn.split('.')

        for label in labels:
            if not self._is_valid_ldh_syntax(label):
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_INVALID_DNSNAME_SYNTAX,
                    f'Non-LDH domain label "{label}" in domain name "{value}"'
                )

            m = self._ACE_REGEX.search(label)

            if m is None:
                continue

            tag = m.group('tag')

            if tag.casefold() != 'xn'.casefold():
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_PROHIBITED_RESERVED_LABEL,
                    f'Invalid reserved label "{label}" with tag "{tag}" in domain name "{value}"'
                )

            try:
                _ = m.group('ace').encode().decode('punycode')
            except UnicodeError:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_INVALID_PLABEL,
                    f'Invalid P-Label "{label}" in domain name "{value}"'
                )


class TorVersion3DomainNameValidator(validation.Validator):
    """Validates that each Onion Domain Name conforms to the Tor v3 specification."""

    VALIDATION_INVALID_TOR_V3_NAME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.invalid_tor_v3_domain_name'
    )

    VALIDATION_INVALID_TOR_VERSION = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.invalid_tor_version'
    )

    VALIDATION_INVALID_TOR_CHECKSUM = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.invalid_tor_checksum'
    )

    _CHECKSUM = b'.onion checksum'
    _VERSION = 0x03

    _ONION_V3_DOMAIN_NAME_REGEX = re.compile(r'^([^.]+\.)*(?P<descriptor>[a-z2-7]{56})(\.onion)?$', re.IGNORECASE)

    def __init__(self):
        super().__init__(validations=[
                            self.VALIDATION_INVALID_TOR_V3_NAME,
                            self.VALIDATION_INVALID_TOR_VERSION,
                            self.VALIDATION_INVALID_TOR_CHECKSUM],
                         predicate=general_name.create_generalname_type_predicate('dNSName'))

    @staticmethod
    def _calculate_checksum(pubkey_octets):
        s = bytearray()
        s.extend(TorVersion3DomainNameValidator._CHECKSUM)
        s.extend(pubkey_octets)
        s.append(TorVersion3DomainNameValidator._VERSION)

        h = hashes.Hash(hashes.SHA3_256())
        h.update(s)
        digest = h.finalize()

        return digest[:2]

    def validate(self, node):
        value = str(node.pdu)

        if not value.lower().endswith('.onion'):
            return

        m = self._ONION_V3_DOMAIN_NAME_REGEX.match(value)

        if m is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_TOR_V3_NAME,
                f'Invalid Tor v3 domain name: "{value}"'
            )

        try:
            decoded = base64.b32decode(m.group('descriptor').upper())
        except binascii.Error:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_TOR_V3_NAME,
                f'Invalid Base-32 encoding for descriptor: "{value}"'
            )

        version = decoded[-1]

        if version != self._VERSION:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_TOR_VERSION,
                f'Invalid Tor version {version} in domain name: "{value}"'
            )

        actual_checksum = decoded[-3:-1]
        pubkey_octets = decoded[0:32]

        expected_checksum = TorVersion3DomainNameValidator._calculate_checksum(pubkey_octets)

        if actual_checksum != expected_checksum:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_TOR_CHECKSUM,
                f'Invalid Tor v3 checksum. Expected: {expected_checksum.hex()}, actual: {actual_checksum.hex()}'
            )



import io
import operator
from typing import Iterable

from pkilint import document, loader, validation


def create_document(pdu):
    name = document.get_node_name_for_pdu(pdu)
    doc = document.Document(None, 'test', pdu)

    doc.root = document.PDUNode(doc, name, pdu, None)

    return doc.root


class ExpectedResult(document.NodeVisitor):
    def __init__(
            self,
            expected_findings,
            expected_count=1,
            **kwargs
    ):
        self._expected_findings = set(expected_findings)
        self.expected_count = expected_count

        super().__init__(**kwargs)

    def validate(self, actual_result: validation.ValidationResult):
        actual_findings = set(
            map(
                operator.attrgetter('finding'),
                actual_result.finding_descriptions
            )
        )

        assert actual_findings == self._expected_findings, (
            'Actual findings are different from expected findings'
        )

    def __repr__(self) -> str:
        findings_str = '\n'.join([f'  - {f}' for f in self._expected_findings])
        return f'Expected findings: \n{findings_str}\nExpected count: {self.expected_count}'


def compare_results(
        actual_results: Iterable[validation.ValidationResult],
        expected_results=None
):
    if expected_results is None:
        expected_results = []
    for expected_result in expected_results:
        matched = [r for r in actual_results if expected_result.match(r.node)]

        assert len(matched) == expected_result.expected_count, (
            'Unexpected number of node matches. '
            f'Expected: {expected_result.expected_count}  '
            f'Actual: {len(matched)}'
        )

        for actual_result in matched:
            expected_result.validate(actual_result)

    for result_with_finding in (r for r in actual_results if any(r.finding_descriptions)):
        expected = False
        for expected_result in expected_results:
            if expected_result.match(result_with_finding.node):
                expected = True
                break

        assert expected, (
            f'Result has unexpected findings: {result_with_finding}'
        )


def certificate_test_harness(
        substrate,
        validator: validation.Validator,
        expected_results=None,
        decoding_container: validation.ValidatorContainer = None
):
    if expected_results is None:
        expected_results = []
    if isinstance(substrate, str):
        substrate = substrate.encode('us-ascii')

    doc = loader.load_certificate(io.BytesIO(substrate), 'test')

    doc_validator_container = []
    if decoding_container is not None:
        doc_validator_container.append(decoding_container)

    doc_validator_container.append(validator)

    container = validation.ValidatorContainer(
        validators=doc_validator_container
    )

    results = container.validate(doc.root)

    compare_results(results, expected_results)



from pyasn1_alt_modules import rfc5280, rfc6962

import pkilint.common
from pkilint import validation, oid, common, document
from pkilint.cabf.serverauth import serverauth_constants
from pkilint.itu import x520_name
from pkilint.pkix import Rfc2119Word


class CaCertificatePoliciesValidator(validation.Validator):
    """Validates that the content of the certificate policies extension complies with BR  7.1.2.10.5."""

    VALIDATION_ANYPOLICY_EXTERNAL_CA = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.ca_external_anypolicy'
    )

    VALIDATION_MULTIPLE_RESERVED_OIDS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.ca_multiple_reserved_policy_oids'
    )

    VALIDATION_NO_RESERVED_OID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.ca_missing_reserved_policy_oid'
    )

    VALIDATION_NON_TLS_CA_HAS_SERVERAUTH_OID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.ca_non_tls_has_reserved_policy_oid'
    )

    VALIDATION_ANYPOLICY_WITH_OTHER_OID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.ca_anypolicy_with_other_policy_oid'
    )

    VALIDATION_FIRST_OID_NOT_RESERVED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.ca_first_policy_oid_not_reserved'
    )

    def __init__(self, certificate_type: serverauth_constants.CertificateType):
        self._certificate_type = certificate_type

        super().__init__(validations=[self.VALIDATION_ANYPOLICY_EXTERNAL_CA, self.VALIDATION_MULTIPLE_RESERVED_OIDS,
                                      self.VALIDATION_NO_RESERVED_OID, self.VALIDATION_NON_TLS_CA_HAS_SERVERAUTH_OID,
                                      self.VALIDATION_ANYPOLICY_WITH_OTHER_OID, self.VALIDATION_FIRST_OID_NOT_RESERVED],
                         pdu_class=rfc5280.CertificatePolicies)

    def validate(self, node):
        policy_oids = [pi.children['policyIdentifier'].pdu for pi in node.children.values()]

        has_any_policy = rfc5280.anyPolicy in policy_oids

        if has_any_policy:
            if self._certificate_type in serverauth_constants.EXTERNAL_CA_TYPES:
                raise validation.ValidationFindingEncountered(self.VALIDATION_ANYPOLICY_EXTERNAL_CA)

            if len(policy_oids) > 1:
                raise validation.ValidationFindingEncountered(self.VALIDATION_ANYPOLICY_WITH_OTHER_OID)
        else:
            reserved_oids = set(policy_oids) & serverauth_constants.SERVERAUTH_RESERVED_POLICY_OIDS

            if self._certificate_type == serverauth_constants.CertificateType.NON_TLS_CA:
                if any(reserved_oids):
                    oids = oid.format_oids(reserved_oids)

                    raise validation.ValidationFindingEncountered(
                        self.VALIDATION_NON_TLS_CA_HAS_SERVERAUTH_OID,
                        f'Non-TLS CA has reserved policy OIDs: {oids}'
                    )
            else:
                if not any(reserved_oids):
                    raise validation.ValidationFindingEncountered(self.VALIDATION_NO_RESERVED_OID)

            if len(reserved_oids) > 1:
                oids_str = oid.format_oids(reserved_oids)

                raise validation.ValidationFindingEncountered(self.VALIDATION_MULTIPLE_RESERVED_OIDS,
                                                              f'Multiple reserved policy OIDs present: {oids_str}')

            if (self._certificate_type != serverauth_constants.CertificateType.NON_TLS_CA and
                    policy_oids[0] not in serverauth_constants.SERVERAUTH_RESERVED_POLICY_OIDS):
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_FIRST_OID_NOT_RESERVED
                )


class TlsCaCertificateAllowedEkuValidator(common.ExtendedKeyUsageAllowanceValidator):
    """Validates that the content of the extended key usage extension complies with BR  7.1.2.10.6."""

    _CODE_CLASSIFIER = 'cabf.serverauth.ca'

    _EKU_ALLOWANCES = {
        **{e: Rfc2119Word.MUST_NOT for e in (
            rfc5280.id_kp_codeSigning,
            rfc5280.id_kp_emailProtection,
            rfc5280.id_kp_OCSPSigning,
            rfc5280.anyExtendedKeyUsage,
            rfc6962.id_kp_precertificateSigning
        )},
        rfc5280.id_kp_serverAuth: Rfc2119Word.MUST,
        rfc5280.id_kp_clientAuth: Rfc2119Word.MAY,
    }

    def __init__(self):
        super().__init__(self._EKU_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


# BR 7.1.2.4.2
class PrecertSigningCaCertificateAllowedEkuValidator(common.ExtendedKeyUsageAllowanceValidator):
    """Validates that the content of the extended key usage extension complies with BR  7.1.2.4.2."""

    _CODE_CLASSIFIER = 'cabf.serverauth.ca_precert_signing'

    _EKU_ALLOWANCES = {
        rfc6962.id_kp_precertificateSigning: Rfc2119Word.MUST,
    }

    def __init__(self):
        super().__init__(self._EKU_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MUST_NOT)


# BR 7.1.2.3.3
class NonTlsCaCertificateAllowedEkuValidator(common.ExtendedKeyUsageAllowanceValidator):
    _CODE_CLASSIFIER = 'cabf.serverauth.non_tls_ca'

    _EKU_ALLOWANCES = {
        e: Rfc2119Word.MUST_NOT for e in (
            rfc5280.id_kp_serverAuth,
            rfc5280.id_kp_OCSPSigning,
            rfc5280.anyExtendedKeyUsage,
            rfc6962.id_kp_precertificateSigning,
        )
    }

    def __init__(self):
        super().__init__(self._EKU_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MAY)


class CaRequiredSubjectAttributesValidator(pkilint.common.AttributeIdentifierAllowanceValidator):
    """Validates that the subject contains attributes in accordance with BR 7.1.2.10.2."""

    _CODE_CLASSIFIER = 'cabf.serverauth.ca'

    _ATTRIBUTE_ALLOWANCES = {
        rfc5280.id_at_countryName: Rfc2119Word.MUST,
        rfc5280.id_at_stateOrProvinceName: Rfc2119Word.MAY,
        rfc5280.id_at_localityName: Rfc2119Word.MAY,
        x520_name.id_at_postalCode: Rfc2119Word.MAY,
        x520_name.id_at_streetAddress: Rfc2119Word.MAY,
        rfc5280.id_at_organizationName: Rfc2119Word.MUST,
        rfc5280.id_at_commonName: Rfc2119Word.MUST,
    }

    def __init__(self, certificate_type: serverauth_constants.CertificateType):
        self._attribute_allowances = self._ATTRIBUTE_ALLOWANCES.copy()

        if certificate_type in (serverauth_constants.TLS_CA_TYPES | {serverauth_constants.CertificateType.ROOT_CA}):
            ou_allowance_word = Rfc2119Word.MUST_NOT
        else:
            ou_allowance_word = Rfc2119Word.SHOULD_NOT

        self._attribute_allowances[rfc5280.id_at_organizationalUnitName] = ou_allowance_word

        super().__init__(self._attribute_allowances, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


class NameConstraintsBaseTypeValidator(validation.Validator):
    """Validates that each subtree of a name constraints extension conforms with BR 7.1.2.10.8."""

    VALIDATION_DISCOURAGED_BASE_NAME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.name_constraints_discouraged_name_type'
    )

    VALIDATION_DIRNAME_IN_EXCLUDED_SUBTREES = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.serverauth.name_constraints_dirname_in_excluded_subtrees'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_DISCOURAGED_BASE_NAME_TYPE,
                                      self.VALIDATION_DIRNAME_IN_EXCLUDED_SUBTREES],
                         pdu_class=rfc5280.GeneralSubtree)

    def validate(self, node):
        gn_type, gn_value = node.children['base'].child

        if gn_type not in {'dNSName', 'iPAddress', 'directoryName'}:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_DISCOURAGED_BASE_NAME_TYPE,
                f'Discouraged GeneralTree base type: {gn_type}'
            )

        if node.parent.name == 'excludedSubtrees' and gn_type == 'directoryName':
            raise validation.ValidationFindingEncountered(self.VALIDATION_DIRNAME_IN_EXCLUDED_SUBTREES)


# all extension criticalities are aligned for the CA profiles; BR 7.1.2.5.1 has the most comprehensive specification
class CaCertificateExtensionCriticalityValidator(common.ExtensionCriticalityValidator):
    """Validates that the criticality of all extensions conforms with BR 7.1.2.5.1."""

    _CODE_CLASSIFIER = 'cabf.serverauth.ca'

    _CRITICALITY_MAPPING = {
        rfc5280.id_ce_authorityKeyIdentifier: False,
        rfc5280.id_ce_basicConstraints: True,
        rfc5280.id_ce_certificatePolicies: False,
        rfc5280.id_ce_cRLDistributionPoints: False,
        rfc5280.id_ce_keyUsage: True,
        rfc5280.id_ce_subjectKeyIdentifier: False,
        rfc5280.id_ce_extKeyUsage: False,
        rfc5280.id_pe_authorityInfoAccess: False,
        rfc6962.id_ce_embeddedSCT: False,
    }

    def __init__(self):
        super().__init__(self._CRITICALITY_MAPPING, self._CODE_CLASSIFIER, Rfc2119Word.MUST, Rfc2119Word.MUST)


class CaCertificateExtensionAllowanceValidator(common.ExtensionIdentifierAllowanceValidator):
    """Validates that the included extensions conform with BR 7.1.2.4.1, 7.1.2.5.1, or 7.1.2.6.1 (depending on
    certificate type)"""

    _CODE_CLASSIFIER = 'cabf.serverauth.ca'

    _EXTENSION_ALLOWANCES = {
        rfc5280.id_ce_authorityKeyIdentifier: Rfc2119Word.MUST,
        rfc5280.id_ce_basicConstraints: Rfc2119Word.MUST,
        rfc5280.id_ce_certificatePolicies: Rfc2119Word.MUST,
        rfc5280.id_ce_cRLDistributionPoints: Rfc2119Word.MUST,
        rfc5280.id_ce_keyUsage: Rfc2119Word.MUST,
        rfc5280.id_ce_subjectKeyIdentifier: Rfc2119Word.MUST,
        rfc5280.id_ce_extKeyUsage: Rfc2119Word.MUST,
        rfc5280.id_pe_authorityInfoAccess: Rfc2119Word.SHOULD,
    }

    def __init__(self, certificate_type):
        extension_allowances = self._EXTENSION_ALLOWANCES.copy()

        if certificate_type in serverauth_constants.CONSTRAINED_TLS_CA_TYPES:
            extension_allowances[rfc5280.id_ce_nameConstraints] = Rfc2119Word.MUST
        else:
            extension_allowances[rfc5280.id_ce_nameConstraints] = Rfc2119Word.MAY

        if certificate_type != serverauth_constants.CertificateType.PRECERT_SIGNING_CA:
            extension_allowances[rfc6962.id_ce_embeddedSCT] = Rfc2119Word.MAY

        super().__init__(extension_allowances, self._CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


class CaCertificateAuthorityInformationAccessAccessMethodPresenceValidator(
        common.AuthorityInformationAccessAccessMethodPresenceValidator):
    """Validates that the content of the authority information access extension conforms to BR 7.1.2.10.3."""

    _CODE_CLASSIFIER = 'cabf.serverauth.ca'

    _ACCESS_METHOD_ALLOWANCES = {
        rfc5280.id_ad_ocsp: Rfc2119Word.SHOULD,
        rfc5280.id_ad_caIssuers: Rfc2119Word.MAY,
    }

    def __init__(self):
        super().__init__(self._ACCESS_METHOD_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MUST_NOT)


# BR 7.1.2.5.2
class TlsCaTechnicallyConstrainedValidator(validation.Validator):
    """Validates that the CA is technically constrained in accordance with BR 7.1.2.5.2."""

    VALIDATION_INCOMPLETE_NAME_CONSTRAINTS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.ca.incomplete_name_constraints'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_INCOMPLETE_NAME_CONSTRAINTS, pdu_class=rfc5280.NameConstraints)

    def validate(self, node):
        try:
            permitted_subtrees = node.navigate('permittedSubtrees')
        except document.PDUNavigationFailedError:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INCOMPLETE_NAME_CONSTRAINTS,
                '"permittedSubtrees" is absent'
            )

        has_dirname_constraint = False
        has_dnsname_constraint = False
        has_ipv4_constraint = False
        has_ipv6_constraint = False

        for subtree in permitted_subtrees.children.values():
            gn_type, gn_value = subtree.children['base'].child

            if gn_type == 'directoryName':
                has_dirname_constraint = True
            elif gn_type == 'dNSName':
                has_dnsname_constraint = True
            elif gn_type == 'iPAddress':
                ip_network_octet_len = len(gn_value.pdu.asOctets())

                if ip_network_octet_len == 8:
                    has_ipv4_constraint = True
                else:
                    has_ipv6_constraint = True

        if not has_dirname_constraint:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INCOMPLETE_NAME_CONSTRAINTS,
                '"permittedSubtrees" does not contain a directoryName'
            )

        if has_dirname_constraint and has_dnsname_constraint and has_ipv4_constraint and has_ipv6_constraint:
            return

        try:
            excluded_subtrees = node.navigate('excludedSubtrees')
        except document.PDUNavigationFailedError:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INCOMPLETE_NAME_CONSTRAINTS,
                '"excludedSubtrees" absent with incomplete constraints in "permittedSubtrees"'
            )

        for subtree in excluded_subtrees.children.values():
            gn_type, gn_value = subtree.children['base'].child

            if gn_type == 'dNSName' and len(str(gn_value.pdu)) == 0:
                has_dnsname_constraint = True
            elif gn_type == 'iPAddress':
                ip_network_octets = gn_value.pdu.asOctets()

                if all((o == 0 for o in ip_network_octets)):
                    ip_network_octet_len = len(ip_network_octets)

                    if ip_network_octet_len == 8:
                        has_ipv4_constraint = True
                    else:
                        has_ipv6_constraint = True

        if has_dirname_constraint and has_dnsname_constraint and has_ipv4_constraint and has_ipv6_constraint:
            return
        else:
            raise validation.ValidationFindingEncountered(self.VALIDATION_INCOMPLETE_NAME_CONSTRAINTS)



from importlib.metadata import version
from typing import List

from fastapi import FastAPI, HTTPException

from pkilint.rest import cabf_serverauth, cabf_smime
from pkilint.rest import model

_PKILINT_VERSION = version('pkilint')
_API_VERSION = 'v1'

app = FastAPI(
    title='pkilint API',
    version=_API_VERSION,
    description='HTTP interface for pkilint'
)


_CERTIFICATE_LINTER_GROUPS = [
    cabf_smime.create_linter_group_instance(),
    cabf_serverauth.create_linter_group_instance(),
]


@app.get('/version')
def version() -> model.Version:
    """Retrieves the version of pkilint that this server is using"""
    return model.Version(version=_PKILINT_VERSION)


@app.get('/certificate')
def certificate_linter_groups() -> List[model.LinterGroup]:
    """Retrieves the groups of linters that are available for linting certificates"""
    return _CERTIFICATE_LINTER_GROUPS


def _get_linter_group_by_name(linter_group_name: str):
    try:
        return next((l for l in _CERTIFICATE_LINTER_GROUPS if l.name.casefold() == linter_group_name.casefold()))
    except StopIteration:
        raise HTTPException(404, 'Linter group with the specified name does not exist')


@app.get('/certificate/{linter_group_name}')
def certificate_linter_group(linter_group_name: str) -> model.LinterGroup:
    """Retrieves the list of linters that are available in the specified linter group"""
    return _get_linter_group_by_name(linter_group_name)


@app.post('/certificate/{linter_group_name}')
def certificate_determine_and_lint(
        linter_group_name: str, doc: model.CertificateModel) -> model.LintResultListWithLinter:
    """Determines the linter that is most appropriate to lint the specified certificate and then returns the results
    reported by the linter for the certificate"""

    linter_group_instance = _get_linter_group_by_name(linter_group_name)

    parsed_doc = doc.parsed_document

    linter = linter_group_instance.determine_linter(parsed_doc)

    result_list = linter.lint(parsed_doc)

    return model.LintResultListWithLinter(results=result_list.results, linter=linter)


@app.post('/certificate/{linter_group_name}/determine-linter')
def certificate_determine_type(linter_group_name: str, doc: model.CertificateModel) -> model.Linter:
    """Determines the linter that is most appropriate to lint the specified certificate"""
    linter_group_instance = _get_linter_group_by_name(linter_group_name)

    parsed_doc = doc.parsed_document

    return linter_group_instance.determine_linter(parsed_doc)


@app.post('/certificate/{linter_group_name}/{linter_name}')
def certificate_lint(linter_group_name: str, linter_name: str, doc: model.CertificateModel) -> model.LintResultList:
    """Lints the specified certificate with the specified linter"""
    linter_group_instance = _get_linter_group_by_name(linter_group_name)

    linter_instance = linter_group_instance.get_linter_by_name(linter_name)

    parsed_doc = doc.parsed_document

    return linter_instance.lint(parsed_doc)



from pyasn1_alt_modules import rfc5280

from pkilint import validation, common
from pkilint.pkix import extension


class AuthorityInformationAccessPresenceValidator(extension.ExtensionPresenceValidator):
    _VALIDATION_TYPE = 'cabf.aia_extension_missing'

    def __init__(self, severity):
        finding = validation.ValidationFinding(
            severity,
            self._VALIDATION_TYPE
        )

        super().__init__(extension_oid=rfc5280.id_pe_authorityInfoAccess,
                         validation=finding,
                         pdu_class=rfc5280.Extensions
                         )


def _general_name_is_http_uri(node):
    name, value = node.child

    return (
            name == 'uniformResourceIdentifier' and
            str(value.pdu).lower().startswith('http:')
    )


def _validate_descriptions(node, access_method, empty_finding,
                           no_http_finding
                           ):
    ads = [
        ad
        for ad in node.children.values()
        if ad.children['accessMethod'].pdu == access_method
    ]

    if len(ads) == 0:
        if empty_finding is None:
            return
        else:
            raise validation.ValidationFindingEncountered(empty_finding)

    if not any(
            filter(
                lambda a: _general_name_is_http_uri(a.children['accessLocation']),
                ads
            )
    ):
        raise validation.ValidationFindingEncountered(no_http_finding)


class AuthorityInformationAccessContainsHttpUriValidator(validation.Validator):
    VALIDATION_NO_CA_ISSUERS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.aia_ca_issuers_missing'
    )

    VALIDATION_CA_ISSUERS_NO_HTTP_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.aia_ca_issuers_has_no_http_uri'
    )

    VALIDATION_OCSP_NO_HTTP_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.aia_ocsp_has_no_http_uri'
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_NO_CA_ISSUERS,
                self.VALIDATION_CA_ISSUERS_NO_HTTP_URI,
                self.VALIDATION_OCSP_NO_HTTP_URI,
            ],
            pdu_class=rfc5280.AuthorityInfoAccessSyntax
        )

    def validate(self, node):
        _validate_descriptions(
            node, rfc5280.id_ad_caIssuers, self.VALIDATION_NO_CA_ISSUERS,
            self.VALIDATION_CA_ISSUERS_NO_HTTP_URI
        )
        _validate_descriptions(
            node, rfc5280.id_ad_ocsp, None,
            self.VALIDATION_OCSP_NO_HTTP_URI
        )


class CrlDpContainsHttpUriValidator(validation.Validator):
    VALIDATION_CRLDP_NO_HTTP_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.no_http_crldp_uri'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_CRLDP_NO_HTTP_URI],
            pdu_class=rfc5280.CRLDistributionPoints
        )

    def validate(self, node):
        for dist_point in node.children.values():
            dp = dist_point.children.get('distributionPoint')

            if dp is None:
                continue

            full_name = dp.children.get('fullName')
            if full_name is None:
                continue

            for gn in full_name.children.values():
                name, value = gn.child
                if (
                        name == 'uniformResourceIdentifier' and
                        str(value.pdu).lower().startswith('http:')
                ):
                    return

        raise validation.ValidationFindingEncountered(
            self.VALIDATION_CRLDP_NO_HTTP_URI
        )


class CertificatePoliciesCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_CERTIFICATE_POLICIES_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'cabf.critical_certificate_policies_extension'
    )

    def __init__(self):
        super().__init__(validation=self.VALIDATION_CERTIFICATE_POLICIES_CRITICAL,
                         type_oid=rfc5280.id_ce_certificatePolicies,
                         is_critical=False)


class CabfCrlDpCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_CRLDP_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.critical_crldp_extension'
    )

    def __init__(self):
        super().__init__(validation=self.VALIDATION_CRLDP_CRITICAL,
                         type_oid=rfc5280.id_ce_cRLDistributionPoints,
                         is_critical=False)


class CabfAuthorityKeyIdentifierValidator(validation.Validator):
    VALIDATION_AKI_HAS_ISSUER_CERT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.authority_key_identifier_has_issuer_cert'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_AKI_HAS_ISSUER_CERT],
            pdu_class=rfc5280.AuthorityKeyIdentifier
        )

    def validate(self, node):
        if (
                'authorityCertIssuer' in node.children or
                'authorityCertSerialNumber' in node.children
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AKI_HAS_ISSUER_CERT
            )


class CabfExtensionsPresenceValidator(common.ExtensionsPresenceValidator):
    VALIDATION_EXTENSIONS_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.certificate_extensions_missing'
    )

    def __init__(self):
        super().__init__(self.VALIDATION_EXTENSIONS_MISSING)


class CpsUriHttpValidator(validation.Validator):
    VALIDATION_CPS_URI_NOT_HTTP = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.cps_uri_is_not_http'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_CPS_URI_NOT_HTTP], pdu_class=rfc5280.CPSuri)

    def validate(self, node):
        uri = str(node.pdu)
        scheme = uri.split(':', maxsplit=1)[0]

        if scheme.lower() not in {'http', 'https'}:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CPS_URI_NOT_HTTP,
                f'Prohibited URI scheme: {scheme}'
            )
